# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2011-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Qt implementation of the UI."""

from functools import wraps

# pylint: disable=F0401,E0611
from PyQt4.QtCore import Qt, pyqtSignal
from PyQt4.QtGui import (
    QApplication,
    QCursor,
    QFrame,
    QHBoxLayout,
    QLabel,
    QStyle,
    QVBoxLayout,
    QWizardPage,
)
from twisted.internet import defer

from ubuntu_kylin_sso import main
from ubuntu_kylin_sso.constants import APP_NAME, TC_URL, POLICY_URL
from ubuntu_kylin_sso.logger import setup_gui_logging, log_call
from ubuntu_kylin_sso.qt import (

    ERROR_STYLE,
    maybe_elide_text,
    PREFERED_UI_SIZE,
    TITLE_STYLE,
)
from ubuntu_kylin_sso.utils.ui import GENERIC_BACKEND_ERROR


logger = setup_gui_logging('ubuntu_kylin_sso.sso_wizard_page')


class WizardHeader(QFrame):
    """WizardHeader Class for Title and Subtitle in all wizard pages."""

    def __init__(self, max_width, parent=None):
        """Create a new instance."""
        super(WizardHeader, self).__init__(parent=parent)
        self.max_width = max_width
        self.max_title_width = self.max_width * 0.95
        self.max_subtitle_width = self.max_width * 1.8

        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(0, 0, 0, 0)
        self.title_label = QLabel()
        self.title_label.setWordWrap(True)
        self.title_label.setObjectName('title_label')
        self.subtitle_label = QLabel()
        self.subtitle_label.setWordWrap(True)
        self.subtitle_label.setFixedHeight(32)
        vbox.addWidget(self.title_label)
        vbox.addWidget(self.subtitle_label)
        self.title_label.setVisible(False)
        self.subtitle_label.setVisible(False)

    def set_title(self, title):
        """Set the Title of the page or hide it otherwise"""
        if title:
            maybe_elide_text(self.title_label, title, self.max_title_width,
                             markup=TITLE_STYLE)
            self.title_label.setVisible(True)
        else:
            self.title_label.setVisible(False)

    def set_subtitle(self, subtitle):
        """Set the Subtitle of the page or hide it otherwise"""
        if subtitle:
            maybe_elide_text(self.subtitle_label, subtitle,
                             self.max_subtitle_width)
            self.subtitle_label.setVisible(True)
        else:
            self.subtitle_label.setVisible(False)


class BaseWizardPage(QWizardPage):
    """Base class for all wizard pages."""

    ui_class = None
    max_width = 0
    processingStarted = pyqtSignal()
    processingFinished = pyqtSignal()

    def __init__(self, parent=None):
        super(BaseWizardPage, self).__init__(parent=parent)

        self.ui = None
        if self.ui_class is not None:
            # self.ui_class is not callable, pylint: disable=E1102
            self.ui = self.ui_class()
            self.ui.setupUi(self)

        if self.layout() is None:
            self.setLayout(QVBoxLayout(self))

        # Set the error area
        self.form_errors_label = QLabel()
        self.form_errors_label.setObjectName('form_errors')
        self.form_errors_label.setAlignment(Qt.AlignBottom)
        self.layout().insertWidget(0, self.form_errors_label)

        # Set the header
        self.header = WizardHeader(max_width=self.max_width)
        self.header.set_title(title='')
        self.header.set_subtitle(subtitle='')
        self.layout().insertWidget(0, self.header)

        self.layout().setAlignment(Qt.AlignLeft)

        self._is_processing = False

    def _get_is_processing(self):
        """Is this widget processing any request?"""
        return self._is_processing

    def _set_is_processing(self, new_value):
        """Set this widget to be processing a request."""
        self._is_processing = new_value
        self.setEnabled(not new_value)
        if not self._is_processing:
            self.processingFinished.emit()
        else:
            self.processingStarted.emit()

    is_processing = property(fget=_get_is_processing, fset=_set_is_processing)

    # pylint: disable=C0103

    def cleanupPage(self):
        """Hide the errors."""
        self.hide_error()

    def setTitle(self, title=''):
        """Set the Wizard Page Title."""
        self.header.set_title(title)

    def setSubTitle(self, subtitle=''):
        """Set the Wizard Page Subtitle."""
        self.header.set_subtitle(subtitle)

    def title(self):
        """Return the header's title."""
        return self.header.title_label.text()

    def subTitle(self):
        """Return the header's subtitle."""
        return self.header.subtitle_label.text()

    # pylint: enable=C0103

    @log_call(logger.error)
    def show_error(self, message):
        """Show an error message inside the page."""
        self.is_processing = False
        maybe_elide_text(self.form_errors_label, message,
                         self.max_width * 0.95, markup=ERROR_STYLE)

    def hide_error(self):
        """Hide the label errors in the current page."""
        # We actually want the label with one empty char, because if it is an
        # empty string, the height of the label is 0
        self.form_errors_label.setText(' ')


class SSOWizardPage(BaseWizardPage):
    """Root class for all SSO specific wizard pages."""

    _signals = {}  # override in children
    max_width = PREFERED_UI_SIZE['width']

    def __init__(self, app_name, **kwargs):
        """Create a new instance."""
        parent = kwargs.pop('parent', None)
        super(SSOWizardPage, self).__init__(parent=parent)

        # store common useful data provided by the app
        # if not app_name.strip():
        #     app_name = APP_NAME
        # self.app_name = app_name  # APP_NAME
        self.app_name = APP_NAME
        self.tc_url = TC_URL
        self.policy_url = POLICY_URL
        self.help_text = kwargs.get('help_text', '')

        self._signals_receivers = {}
        self.backend = None

        self.setup_page()

    def hide_overlay(self):
        """Emit the signal to notify the upper container that ends loading."""
        self.is_processing = False

    def show_overlay(self):
        """Emit the signal to notify the upper container that is loading."""
        self.is_processing = True

    @defer.inlineCallbacks
    def setup_page(self):
        """Setup the widget components."""
        logger.info('Starting setup_page for: %r', self)
        # pylint: disable=W0702,W0703
        try:
            # Get Backend
            client = yield main.get_sso_client()
            self.backend = client.sso_login
            self._set_translated_strings()
            self._connect_ui()
            # Call _setup_signals at the end, so we ensure that the UI
            # is at least styled as expected if the operations with the
            # backend fails.
            self._setup_signals()
        except:
            message = 'There was a problem trying to setup the page %r' % self
            self.show_error(message)
            logger.exception(message)
            self.setEnabled(False)
        # pylint: enable=W0702,W0703
        logger.info('%r - setup_page ends, backend is %r.', self, self.backend)

    def _filter_by_app_name(self, f):
        """Excecute the decorated function only for 'self.app_name'."""

        @wraps(f)
        def inner(app_name, *args, **kwargs):
            """Execute 'f' only if 'app_name' matches 'self.app_name'."""
            result = None
            if app_name == self.app_name:
                result = f(app_name, *args, **kwargs)
            else:
                logger.info('%s: ignoring call since received app_name '
                            '"%s" (expected "%s")',
                            f.__name__, app_name, self.app_name)
            return result

        return inner

    def _setup_signals(self):
        """Bind signals to callbacks to be able to test the pages."""
        for signal, method in self._signals.items():
            actual = self._signals_receivers.get(signal)
            if actual is not None:
                msg = 'Signal %r is already connected with %r.'
                logger.warning(msg, signal, actual)

            match = self.backend.connect_to_signal(signal, method)
            self._signals_receivers[signal] = match

    def _set_translated_strings(self):
        """Implement in each child."""

    def _connect_ui(self):
        """Implement in each child."""

    def _handle_error(self, remote_call, handler, error):
        """Handle any error when calling the remote backend."""
        logger.error('Remote call %r failed with: %r', remote_call, error)
        errordict = {'message': GENERIC_BACKEND_ERROR}
        handler(self.app_name, errordict)


class EnhancedLineEdit(object):
    """Represents and enhanced lineedit.

    This class works on an already added lineedit to the widget so
    that we are just adding extra items to it.
    """

    def __init__(self, line_edit, valid_cb=lambda x: False,
                 warning_sign=False):
        """Create an instance."""
        super(EnhancedLineEdit, self).__init__()
        self._line_edit = line_edit
        layout = QHBoxLayout(self._line_edit)
        layout.setMargin(0)
        self._line_edit.setLayout(layout)
        self.valid_cb = valid_cb
        layout.addStretch()
        self.clear_label = QLabel(self._line_edit)
        self.clear_label.setMargin(2)
        self.clear_label.setProperty("lineEditWarning", True)
        layout.addWidget(self.clear_label)
        self.clear_label.setMinimumSize(16, 16)
        self.clear_label.setVisible(False)
        self.clear_label.setCursor(QCursor(Qt.ArrowCursor))
        if warning_sign:
            icon = QApplication.style().standardIcon(
                QStyle.SP_MessageBoxWarning)
            self.clear_label.setPixmap(icon.pixmap(16, 16))
        # connect the change of text to the cation that will check if the
        # text is valid and if the icon should be shown.
        self._line_edit.textChanged.connect(self.show_button)

    def show_button(self, string):
        """Decide if we show the button or not."""
        if not self.valid_cb(string) and self.clear_label.pixmap() is not None:
            self.clear_label.setVisible(True)
        else:
            self.clear_label.setVisible(False)


class SSOWizardEnhancedEditPage(SSOWizardPage):
    """Page that contains enhanced line edits."""

    # Method '_connect_ui', '_set_translated_strings' is abstract in class
    # 'SSOWizardPage' but is not overridden
    # pylint: disable=W0223

    def __init__(self, *args, **kwargs):
        """Create a new instance."""
        self._enhanced_edits = {}
        super(SSOWizardEnhancedEditPage, self).__init__(*args, **kwargs)

    def set_line_edit_validation_rule(self, edit, cb):
        """Set a new enhanced edit so that we can show an icon."""
        if edit in self._enhanced_edits:
            self._enhanced_edits[edit].valid_cb = cb
        else:
            # create a new enhanced edit
            enhanced_edit = EnhancedLineEdit(edit, cb)
            self._enhanced_edits[edit] = enhanced_edit



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT AN WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Specific tests for the qt implementation."""

from PyQt4.QtNetwork import (
    QSslConfiguration,
    QSslError,
)
from twisted.internet import defer
from ubuntuone.devtools.testcases import TestCase

from ubuntu_kylin_sso.utils.webclient import qtnetwork


class SetupProxyTestCase(TestCase):
    """Test the proxy setup."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(SetupProxyTestCase, self).setUp()
        self.called = []

        self.settings = dict(https=dict(username='user', password='pasword'))
        self.patch(qtnetwork.gsettings, 'get_proxy_settings',
                lambda: self.settings)
        self.patch(qtnetwork, "get_cert_dir", lambda: "")
        self.patch(qtnetwork.glob, "glob", lambda dir: [])

        self.proxy = None

        self.real_build_proxy = qtnetwork.build_proxy

        def build_proxy(settings):
            """Log build_proxy calls."""
            self.called.append(('build_proxy', settings))
            self.proxy = self.real_build_proxy(settings)
            return self.proxy

        self.patch(qtnetwork, 'build_proxy', lambda s: build_proxy(s))

        self.addCleanup(self._clean_webclient_instance)

    def _set_old_platform(self, platform):
        """Set back the platform."""
        qtnetwork.sys.platform = platform

    def _clean_webclient_instance(self):
        """Set the webclient not to have a proxy."""
        qtnetwork.WebClient.proxy_instance = None

    def test_setup_no_certs(self):
        # Ensure WebClient can start even if it finds no certs.
        # It may or may not end up working depending on the certs actually
        # being in the system, but not having them locally shouldn't prevent
        # startup.
        qtnetwork.WebClient()

    def test_setup_unreadable_cert(self):
        # If for some reason a cert of our own is found but can't be read,
        # make sure we raise WebClientError.
        # glob only returns paths that actually exist, but we fake this by
        # passing something that can't be opened.
        self.patch(qtnetwork.glob, "glob", lambda dir: ["asdfasdfasdf"])
        self.assertRaises(qtnetwork.WebClientError, qtnetwork.WebClient)


class SetupLinuxProxyTestCase(SetupProxyTestCase):
    """Test setting up the proxy."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(SetupLinuxProxyTestCase, self).setUp()
        old_platform = qtnetwork.sys.platform
        qtnetwork.sys.platform = 'linux'
        self.addCleanup(self._set_old_platform, old_platform)

    def test_setup_proxy(self):
        """Test setting the proxy."""
        qtnetwork.WebClient()
        self.assertEqual(self.proxy, qtnetwork.WebClient.proxy_instance)
        self.assertIn(('build_proxy', self.settings), self.called)

    def test_setup_instance_present(self):
        """Test when the instance is present."""
        # set the instance and assert we did not reset it
        qtnetwork.WebClient.proxy_instance = 'other_proxy'
        qtnetwork.WebClient()
        self.assertNotEqual(self.proxy, qtnetwork.WebClient.proxy_instance)
        self.assertNotIn(('build_proxy', self.settings), self.called)


class FakeQNetworkReply(object):
    """A fake QNetworkReply."""

    def __init__(self):
        """Initialize this fake."""
        url = qtnetwork.QUrl("http://fake_server")
        self._request = qtnetwork.QNetworkRequest(url)

    def request(self):
        """Return a test network request."""
        return self._request


class SSLErrorsTestCase(TestCase):
    """Tests for when SSL errors happen."""

    def test_handle_ssl_errors(self):
        """Test the _handle_ssl_errors method."""
        result = []
        self.patch(qtnetwork.logger, "error", result.append)
        client = qtnetwork.WebClient()
        reply = FakeQNetworkReply()
        errors = [QSslError()]
        client._handle_ssl_errors(reply, errors)
        self.assertTrue(type(result[0]), unicode)

    def test_nonempty_cert_list(self):
        """Test the Qt CA list has more than our included certs."""
        qtnetwork.WebClient()
        ssl_config = QSslConfiguration.defaultConfiguration()
        ca_certs = ssl_config.caCertificates()
        self.assertTrue(len(ca_certs) > 3)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Timestamp synchronization with the server."""

import time

from twisted.internet import defer

from ubuntu_kylin_sso import SSO_AUTH_BASE_URL
from ubuntu_kylin_sso.logger import setup_logging

logger = setup_logging("ubuntu_kylin_sso.utils.webclient.timestamp")
NOCACHE_HEADERS = {"Cache-Control": "no-cache"}


class TimestampChecker(object):
    """A timestamp that's regularly checked with a server."""

    CHECKING_INTERVAL = 60 * 60  # in seconds
    ERROR_INTERVAL = 30  # in seconds
    SERVER_IRI = SSO_AUTH_BASE_URL

    def __init__(self, webclient_class):
        """Initialize this instance."""
        self.next_check = time.time()
        self.skew = 0
        self.webclient_class = webclient_class

    @defer.inlineCallbacks
    def get_server_date_header(self, server_iri):
        """Get the server date using twisted webclient."""
        webclient = self.webclient_class()
        try:
            response = yield webclient.request(server_iri, method="HEAD",
                                               extra_headers=NOCACHE_HEADERS)
            defer.returnValue(response.headers["Date"][0])
        finally:
            webclient.shutdown()

    @defer.inlineCallbacks
    def get_server_time(self):
        """Get the time at the server."""
        date_string = yield self.get_server_date_header(self.SERVER_IRI)
        # delay import, otherwise a default reactor gets installed
        from twisted.web import http
        timestamp = http.stringToDatetime(date_string)
        defer.returnValue(timestamp)

    @defer.inlineCallbacks
    def get_faithful_time(self):
        """Get an accurate timestamp."""
        local_time = time.time()
        if local_time >= self.next_check:
            try:
                server_time = yield self.get_server_time()
                self.next_check = local_time + self.CHECKING_INTERVAL
                self.skew = server_time - local_time
                logger.debug("Calculated server time skew: %r", self.skew)
            # We just log all exceptions while trying to get the server time
            # pylint: disable=W0703
            except Exception as e:
                logger.debug("Error while verifying server time skew: %r", e)
                self.next_check = local_time + self.ERROR_INTERVAL
        # delay import, otherwise a default reactor gets installed
        from twisted.web import http
        logger.debug("Using corrected timestamp: %r",
                  http.datetimeToString(local_time + self.skew))
        defer.returnValue(int(local_time + self.skew))



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Email Verification page UI."""

from ubuntu_kylin_sso.qt.sso_wizard_page import SSOWizardPage
from ubuntu_kylin_sso.qt.ui.error_message_ui import Ui_ErrorPage
from ubuntu_kylin_sso.utils.ui import ERROR


class ErrorPage(SSOWizardPage):
    """Widget used to show the diff errors."""

    ui_class = Ui_ErrorPage
    next = None

    def _set_translated_strings(self):
        """Set the translated strings."""
        self.ui.error_message_label.setText(ERROR)

    def _connect_ui(self):
        """Connect the buttons to perform actions."""
        self.next = -1



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""A webclient backend that uses twisted.web.client."""

import base64

try:
    # pylint: disable=E0611,F0401
    from urllib.parse import urlparse
    # pylint: enable=E0611,F0401
except ImportError:
    from urlparse import urlparse

from twisted.internet import defer

from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils.webclient.common import (
    BaseWebClient,
    HeaderDict,
    Response,
    UnauthorizedError,
    WebClientError,
)

logger = setup_logging("ubuntu_kylin_sso.utils.webclient.txweb")


class RawResponse(object):
    """A raw response from the webcall."""

    def __init__(self, headers, content, code=200, phrase="OK"):
        """Initialize this response."""
        self.headers = headers
        self.content = content
        self.code = code
        self.phrase = phrase


class WebClient(BaseWebClient):
    """A simple web client that does not support proxies, yet."""

    client_factory = None

    def __init__(self, connector=None, context_factory=None, **kwargs):
        """Initialize this webclient."""
        # delay import, otherwise a default reactor gets installed
        from twisted.web import client

        super(WebClient, self).__init__(**kwargs)

        if connector is None:
            from twisted.internet import reactor
            self.connector = reactor
        else:
            self.connector = connector

        if context_factory is None:
            from twisted.internet import ssl
            self.context_factory = ssl.ClientContextFactory()
        else:
            self.context_factory = context_factory

        # decide which client factory to use
        if WebClient.client_factory is None:
            self.client_factory = client.HTTPClientFactory
        else:
            self.client_factory = WebClient.client_factory

    @defer.inlineCallbacks
    def raw_request(self, method, uri, headers, postdata):
        """Make a raw http request."""
        # Twisted wants headers as bytes, but because of oauthlib, they might
        # be unicodes.  Assume utf-8 and revert the encodings.
        bytes_headers = {}
        for key, value in headers.items():
            if isinstance(key, unicode):
                key = key.encode('utf-8')
            if isinstance(value, unicode):
                value = value.encode('utf-8')
            bytes_headers[key] = value
        headers = bytes_headers

        # delay import, otherwise a default reactor gets installed
        from twisted.web import error

        parsed_url = urlparse(uri)

        # pylint: disable=E1101,E1103
        https = parsed_url.scheme == "https"
        host = parsed_url.netloc.split(":")[0]
        # pylint: enable=E1101,E1103
        if parsed_url.port is None:
            port = 443 if https else 80
        else:
            port = parsed_url.port

        factory = self.client_factory(uri, method=method,
                                       postdata=postdata,
                                       headers=headers,
                                       followRedirect=False)
        # pylint: disable=E1103
        if https:
            self.connector.connectSSL(host, port, factory,
                                      self.context_factory)
        else:
            self.connector.connectTCP(host, port, factory)
        # pylint: enable=E1103

        try:
            content = yield factory.deferred
            response = RawResponse(factory.response_headers, content)
        except error.Error as e:
            response = RawResponse(factory.response_headers, e.response,
                                   int(e.status), e.message)
        defer.returnValue(response)

    @defer.inlineCallbacks
    def request(self, iri, method="GET", extra_headers=None,
                oauth_credentials=None, post_content=None):
        """Get the page, or fail trying."""
        # delay import, otherwise a default reactor gets installed
        from twisted.web import http

        uri = self.iri_to_uri(iri)
        headers = yield self.build_request_headers(uri, method, extra_headers,
                                                   oauth_credentials)

        if self.username and self.password:
            auth = base64.b64encode(self.username + ":" + self.password)
            headers["Authorization"] = "Basic " + auth

        try:
            raw_response = yield self.raw_request(method, uri,
                                                  headers=headers,
                                                  postdata=post_content)
            response_headers = HeaderDict(raw_response.headers)
            if method.lower() != "head":
                response_content = raw_response.content
            else:
                response_content = ""
            if raw_response.code == http.OK:
                defer.returnValue(Response(response_content, response_headers))
            if raw_response.code == http.UNAUTHORIZED:
                raise UnauthorizedError(raw_response.phrase,
                                        response_content)
            raise WebClientError(raw_response.phrase, response_content)
        except WebClientError:
            raise
        except Exception as e:
            raise WebClientError(e.message, e)

    def force_use_proxy(self, settings):
        """Setup this webclient to use the given proxy settings."""
        # No direct proxy support in twisted.web.client



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Network states."""


class NetworkState(object):
    """ A simple class to add a label and make debugging easier. """

    def __init__(self, label="unlabeled"):
        self.label = label

    def __repr__(self):
        return "Network state (%s)" % self.label

# Values returned by the callback
(ONLINE, OFFLINE, UNKNOWN) = (NetworkState("online"),
                              NetworkState("offline"),
                              NetworkState("unknown"))

# Internal NetworkManager State constants
NM_STATE_UNKNOWN = 0
NM_STATE_UNKNOWN_LIST = [NM_STATE_UNKNOWN]
NM_STATE_ASLEEP_OLD = 1
NM_STATE_ASLEEP = 10
NM_STATE_ASLEEP_LIST = [NM_STATE_ASLEEP_OLD,
                        NM_STATE_ASLEEP]
NM_STATE_CONNECTING_OLD = 2
NM_STATE_CONNECTING = 40
NM_STATE_CONNECTING_LIST = [NM_STATE_CONNECTING_OLD,
                            NM_STATE_CONNECTING]
NM_STATE_CONNECTED_OLD = 3
NM_STATE_CONNECTED_LOCAL = 50
NM_STATE_CONNECTED_SITE = 60
NM_STATE_CONNECTED_GLOBAL = 70
# Specifically don't include local and site, as they won't let us get to server
NM_STATE_CONNECTED_LIST = [NM_STATE_CONNECTED_OLD,
                           NM_STATE_CONNECTED_GLOBAL]
NM_STATE_DISCONNECTED_OLD = 4
NM_STATE_DISCONNECTED = 20
# For us, local and site connections are the same as diconnected
NM_STATE_DISCONNECTED_LIST = [NM_STATE_DISCONNECTED_OLD,
                              NM_STATE_DISCONNECTED,
                              NM_STATE_CONNECTED_LOCAL,
                              NM_STATE_CONNECTED_SITE]



# -*- coding: utf-8 -*-
#
# Copyright 2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the locale utilities."""

from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.utils import locale

NO_VALUE = object()


class LocaleTestCase(TestCase):
    """Test case for functions related to locale."""

    def setUp(self):
        d = super(LocaleTestCase, self).setUp()
        self.save_environ("LC_CTYPE")
        self.save_environ("LANG")
        return d

    def save_environ(self, env_var_name):
        """Save the value of an environment variable."""
        old_value = locale.os.environ.get(env_var_name, NO_VALUE)
        self.addCleanup(self.restore_environ, env_var_name, old_value)
        locale.os.environ[env_var_name] = "fake_locale"

    def restore_environ(self, env_var_name, old_value):
        """Restore the value of an environment variable."""
        if old_value is NO_VALUE:
            if env_var_name in locale.os.environ:
                del(locale.os.environ[env_var_name])
        else:
            locale.os.environ[env_var_name] = old_value

    def test_fix_turkish_locale_when_turkish(self):
        """The fix_turkish_locale function skips when no locale set."""
        locale.os.environ["LANG"] = "tr_TR.UTF-8"
        del(locale.os.environ["LC_CTYPE"])
        locale.fix_turkish_locale()
        self.assertEqual(locale.os.environ["LC_CTYPE"], locale.SAFE_LOCALE)

    def test_fix_turkish_locale_when_other(self):
        """The fix_turkish_locale function skips when no locale set."""
        locale.os.environ["LANG"] = "en_EN.UTF-8"
        del(locale.os.environ["LC_CTYPE"])
        locale.fix_turkish_locale()
        self.assertEqual(locale.os.environ.get("LC_CTYPE", NO_VALUE), NO_VALUE)

    def test_fix_turkish_locale_when_LANG_unset(self):
        """The fix_turkish_locale function skips when no locale set."""
        del(locale.os.environ["LANG"])
        del(locale.os.environ["LC_CTYPE"])
        locale.fix_turkish_locale()
        self.assertEqual(locale.os.environ.get("LC_CTYPE", NO_VALUE), NO_VALUE)

    def test_fix_turkish_locale_when_LC_CTYPE_not_turkish(self):
        """The fix is skipped if the LC_CTYPE is already non-turkish."""
        original = "es_ES.UTF-8"
        locale.os.environ["LANG"] = "tr_TR.UTF-8"
        locale.os.environ["LC_CTYPE"] = original
        locale.fix_turkish_locale()
        self.assertEqual(locale.os.environ.get("LC_CTYPE", NO_VALUE), original)

    def test_fix_turkish_locale_when_LC_CTYPE_is_turkish(self):
        """The fix is applied if the LC_CTYPE is turkish."""
        locale.os.environ["LANG"] = "es_ES.UTF-8"
        locale.os.environ["LC_CTYPE"] = "tr_TR.UTF-8"
        locale.fix_turkish_locale()
        self.assertEqual(locale.os.environ.get("LC_CTYPE", NO_VALUE),
                         locale.SAFE_LOCALE)



# -*- coding: utf-8 -*-
# Author: Manuel de la Pena <manuel@canonical.com>
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the ui functions."""

from unittest import TestCase

from ubuntu_kylin_sso.utils.ui import (get_password_strength,
                                 is_min_required_password,
                                 is_correct_email)


class GettextTestCase(TestCase):
    """Test that we don't override builtins with gettext."""

    def test_import(self):
        """Test whether importing ui module defines _ as a builtin."""
        # This module is already imported above, but just to make sure, since
        # top imports may change over time.
        import ubuntu_kylin_sso.utils.ui
        assert(ubuntu_kylin_sso.utils.ui)
        self.assertFalse('_' in __builtins__)


class GetPasswordStrengTestCase(TestCase):
    """Test the function that returns the strength of a password."""

    def test_too_small_password(self):
        """Test the points given to a very small password."""
        password = 'abc'
        self.assertEqual(1, get_password_strength(password))

    def test_small_password(self):
        """Test the points given to a small passwod 4 or more chars."""
        password = 'testtwe'
        self.assertEqual(0, get_password_strength(password))

    def test_eight_chars_password(self):
        """Test the points given to a normal 8 chars password."""
        password = 'abcdabcd'
        self.assertEqual(1, get_password_strength(password))

    def test_eight_chars_and_num(self):
        """Test the points given to a 8 chars password with a num."""
        password = 'abcdabc8'
        self.assertEqual(2, get_password_strength(password))

    def test_eight_chars_low_and_cap(self):
        """Test the points given to a 8 chars password with a capital."""
        password = 'abcdabcD'
        self.assertEqual(2, get_password_strength(password))

    def test_eight_chars_low_canp_num(self):
        """Test the points given to a 8 chars password with capitals & num."""
        password = 'abcdab7D'
        self.assertEqual(3, get_password_strength(password))

    def test_eiqgh_chars_and_special(self):
        """Test the points given to a 8 chars password with special chars."""
        password = 'abcdabc*'
        self.assertEqual(2, get_password_strength(password))

    def test_long_password(self):
        """Test the points goven to a long password."""
        password = 'abcdabcdabcd'
        self.assertEqual(2, get_password_strength(password))

    def test_eleven_chars_and_num(self):
        """Test the points of a loong password with a num."""
        password = 'abcdabcdabcd99'
        self.assertEqual(3, get_password_strength(password))

    def test_eleven_chars_low_cap(self):
        """Test the points of a long password with low and cap."""
        password = 'abcdabcdabcdbABCD'
        self.assertEqual(3, get_password_strength(password))

    def test_eleven_num_low_cap(self):
        """Test the points of a long password with num and diff cap."""
        password = 'ABCDabcdacbd723'
        self.assertEqual(4, get_password_strength(password))

    def test_eleven_num_special(self):
        """Test the point of a long password with a number and special char."""
        password = 'abcdabcdabcd*9'
        self.assertEqual(4, get_password_strength(password))


class IsMinRequiredPasswordTestCase(TestCase):
    """Test the fnction that returns if the password is the min required."""

    def test_no_enough_chars(self):
        """Test a password that does not have enough chars."""
        password = 'Test8'
        self.assertFalse(is_min_required_password(password))

    def test_no_uppercase(self):
        """Test a password that does not have an uppercase."""
        password = 'longenoughtobeapassword8'
        self.assertFalse(is_min_required_password(password))

    def test_no_number(self):
        """Test a password that does not have a number."""
        password = 'longenoughtobeapassworD'
        self.assertFalse(is_min_required_password(password))

    def test_correct_password(self):
        """Test a password that is correct."""
        password = 'TodasLasPaswordPasan88'
        self.assertTrue(is_min_required_password(password))


class IsCorrectEmailAddress(TestCase):
    """Test the is_correct_email funtion."""

    def test_is_correct_email_true(self):
        """Test when the email is correct."""
        email = 'manuel@canonical.com'
        self.assertTrue(is_correct_email(email))

    def test_is_correct_email_false(self):
        """Test when the email is not correct."""
        email = 'manuelcanonical.com'
        self.assertFalse(is_correct_email(email))



# -*- coding: utf-8 -*-
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Credential management utilities.

'Credentials' provides the following fault-tolerant methods:

 * find_credentials
 * clear_credentials
 * store_credentials
 * register
 * login

All the methods return a Deferred that will be fired when the operation was
completed.

For details, please read the Credentials class documentation.

"""

from functools import wraps

from twisted.internet import defer

from ubuntu_kylin_sso import (
    UI_EXECUTABLE_QT,
    USER_CANCELLATION,
    USER_SUCCESS,
)
from ubuntu_kylin_sso.keyring import Keyring
from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils import compat, get_bin_cmd, runner


logger = setup_logging('ubuntu_kylin_sso.credentials')


APP_NAME_KEY = 'app_name'
TC_URL_KEY = 'tc_url'
HELP_TEXT_KEY = 'help_text'
WINDOW_ID_KEY = 'window_id'
PING_URL_KEY = 'ping_url'
POLICY_URL_KEY = 'policy_url'
UI_EXECUTABLE_KEY = 'ui_executable'


class CredentialsError(Exception):
    """Generic credentials error."""


class UserCancellationError(CredentialsError):
    """The user cancelled the process of authentication."""


class UserNotValidatedError(CredentialsError):
    """The user is not validated."""


class GUINotAvailableError(CredentialsError):
    """No user graphical interface is available."""


def handle_failures(msg):
    """Handle failures using 'msg' as error message."""

    def middle(f):
        """Decorate 'f' to catch all errors."""

        @wraps(f)
        @defer.inlineCallbacks
        def inner(self, *a, **kw):
            """Call 'f' within a try-except block.

            If any exception occurs, the exception is logged and re-raised.

            """
            result = None
            try:
                result = yield f(self, *a, **kw)
            except:
                logger.exception('%s (app_name: %s): %s.',
                                 f.__name__, self.app_name, msg)
                raise

            defer.returnValue(result)

        return inner

    return middle


class Credentials(object):
    """Credentials management gateway."""

    def __init__(self, app_name, tc_url=None, help_text='',
                 window_id=0, ping_url=None, policy_url=None,
                 ui_executable=UI_EXECUTABLE_QT):
        """Return a Credentials management object.

        'app_name' is the application name to be displayed in the GUI.

        'tc_url' is the url pointing to Terms & Conditions. If None, no
        TOS agreement will be displayed.

        'help_text' is an explanatory text for the end-users, will be shown
         below the headers.

        'window_id' is the id of the window which will be set as a parent of
         the GUI. If 0, no parent will be set.

        'ping_url' is the url that will be pinged when a user registers/logins
        successfully. The user email will be attached to 'ping_url'.

        'policy_url' is the url pointing to the privacy policy. If None, no
        privacy policy agreement will be displayed.

        When the credentials are retrieved successfully, a dictionary like the
        one below is returned:

            {'token': <value>,
             'token_secret': <value>,
             'consumer_key': <value>,
             'consumer_secret': <value>,
             'name': <the token name, matches "[app_name] @ [host name]">}

        """
        self.app_name = app_name
        self.help_text = help_text
        self.window_id = window_id
        self.ping_url = ping_url
        self.tc_url = tc_url
        self.policy_url = policy_url
        self.ui_executable = ui_executable

    @defer.inlineCallbacks
    def _show_ui(self, login_only):
        """Show the UI and wait for it to finish.

        Upon analyzing returning code from the UI process, emit proper signals
        to the caller.

        The caller can specify a preference for the UI, but if the preferred
        one is not available, the service will try to open any available UI.

        If no GUI is available, GUINotAvailableError will be raised.

        """
        # Now use UbuntuKylin forum to setup account - 20141016 Robert
        if not login_only:
            from ubuntu_kylin_sso.constants import REGISTER_URL
            import webbrowser
            webbrowser.open(REGISTER_URL)
            return

        guis = (self.ui_executable, UI_EXECUTABLE_QT)
        for gui_exe_name in guis:
            try:
                args = get_bin_cmd(gui_exe_name)
            except OSError:
                logger.error('The given UI %r does not exist.',
                             gui_exe_name)
            else:
                break
        else:
            raise GUINotAvailableError('Can not find a GUI to present to the '
                                       'user (tried with %r). Aborting.' %
                                       repr(guis))

        for arg in ('app_name', 'help_text', 'ping_url', 'policy_url',
                    'tc_url', 'window_id'):
            value = getattr(self, arg)
            if value:
                args.append('--%s' % arg)
                if not isinstance(value, compat.basestring):
                    value = compat.text_type(value)
                args.append(value)

        if login_only:
            args.append('--login_only')

        return_code = yield runner.spawn_program(args)
        logger.info('_show_ui: received from the ui return code %r.',
                    return_code)

        credentials = None
        if return_code == USER_SUCCESS:
            credentials = yield self.find_credentials()
        elif return_code == USER_CANCELLATION:
            raise UserCancellationError()
        else:
            raise CredentialsError(return_code)

        defer.returnValue(credentials)

    def _do_login(self, email, password):
        """Login using email/password, connect outcome signals."""
        from ubuntu_kylin_sso.main import SSOLogin

        d = defer.Deferred()

        class DummyProxy(object):
            """A temporary proxy to handle non-traditional login."""

            # pylint: disable=C0103

            def LoggedIn(self, app_name, result):
                """User was logged in."""
                d.callback(result)

            def LoginError(self, app_name, error):
                """There was an error on login."""
                error = CredentialsError(error['errtype'])
                d.errback(error)

            def UserNotValidated(self, app_name, email):
                """User is not validated."""
                d.errback(UserNotValidatedError(email))

            # pylint: enable=C0103

        inner = SSOLogin(proxy=DummyProxy())
        inner.login(app_name=self.app_name,
                    email=email, password=password,
                    ping_url=self.ping_url)

        d.addCallback(lambda _: self.find_credentials())
        return d

    @defer.inlineCallbacks
    def _login_or_register(self, login_only, email=None, password=None):
        """Get credentials if found else prompt the GUI.

        Will return either the credentials, or will raise UserCancellationError
        if the user aborted the operation when the UI was opened.

        """
        logger.info("_login_or_register: login_only=%r email=%r.",
                    login_only, email)
        token = yield self.find_credentials()
        if not token:
            if email and password:
                token = yield self._do_login(email, password)
            else:
                token = yield self._show_ui(login_only)

        defer.returnValue(token)

    @handle_failures(msg='Problem while getting credentials from the keyring')
    @defer.inlineCallbacks
    def find_credentials(self):
        """Get the credentials for 'self.app_name'. Return {} if not there."""
        creds = yield Keyring().get_credentials(self.app_name)
        logger.info('find_credentials: self.app_name %r, '
                    'result is {}? %s', self.app_name, creds is None)
        defer.returnValue(creds if creds is not None else {})

    @handle_failures(msg='Problem while clearing credentials in the keyring')
    def clear_credentials(self):
        """Clear the credentials for 'self.app_name'."""
        return Keyring().delete_credentials(self.app_name)

    @handle_failures(msg='Problem while storing credentials in the keyring')
    def store_credentials(self, token):
        """Store the credentials for 'self.app_name'."""
        return Keyring().set_credentials(self.app_name, token)

    @handle_failures(msg='Problem while performing register')
    def register(self):
        """Get credentials if found else prompt the GUI to register."""
        return self._login_or_register(login_only=False)

    @handle_failures(msg='Problem while performing login')
    def login(self, email=None, password=None):
        """Get credentials if found else prompt the GUI to login.

        if 'email' and 'password' are given, do not prompt the user and use
        that to retrieve a token.

        """
        if email is None or password is None:
            return self._login_or_register(login_only=True)
        else:
            return self._login_or_register(login_only=True,
                                           email=email, password=password)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Pages from SSO."""

from twisted.internet import defer
from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso import networkstate
from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.qt.sso_wizard_page import SSOWizardPage
from ubuntu_kylin_sso.qt.ui import network_detection_ui
from ubuntu_kylin_sso.utils.ui import (
    CLOSE_AND_SETUP_LATER,
    NETWORK_DETECTION_TITLE,
    NETWORK_DETECTION_WARNING,
    TRY_AGAIN_BUTTON,
)


logger = setup_logging('ubuntu_kylin_sso.network_detection_page')


class NetworkDetectionPage(SSOWizardPage):

    """Widget to show if we don't detect a network connection."""

    ui_class = network_detection_ui.Ui_Form
    connectionDetected = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(NetworkDetectionPage, self).__init__(*args, **kwargs)
        banner_pixmap = kwargs.pop('banner_pixmap', None)
        if banner_pixmap is not None:
            self.ui.image_label.setPixmap(banner_pixmap)
        self.btn_try_again = None

    # pylint: disable=C0103

    def initializePage(self):
        """Set UI details."""
        logger.debug('initializePage - About to show NetworkDetectionPage')
        self.wizard()._next_id = -1

        self.setButtonText(QtGui.QWizard.CustomButton1, TRY_AGAIN_BUTTON)
        self.setButtonText(QtGui.QWizard.CancelButton,
            CLOSE_AND_SETUP_LATER)
        self.wizard().setButtonLayout([QtGui.QWizard.Stretch,
                                       QtGui.QWizard.CustomButton1,
                                       QtGui.QWizard.CancelButton,
                                       ])

        try:
            self.wizard().customButtonClicked.disconnect()
        except TypeError:
            pass

        self.ui.label.setText(NETWORK_DETECTION_WARNING %
                              {'app_name': self.app_name})
        self.btn_try_again = self.wizard().button(QtGui.QWizard.CustomButton1)
        self.btn_try_again.setDefault(True)
        self.wizard().customButtonClicked.connect(self.try_again)

    # pylint: enable=C0103

    @defer.inlineCallbacks
    def try_again(self, button_id=QtGui.QWizard.CustomButton1):
        """Test the connection again."""
        if button_id == QtGui.QWizard.CustomButton1:
            d = yield networkstate.is_machine_connected()
            if d:
                self.connectionDetected.emit()

    def _set_translated_strings(self):
        """Implement in each child."""
        self.setTitle(NETWORK_DETECTION_TITLE)

    def _connect_ui(self):
        """Implement in each child."""



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the ssl dialog."""

from PyQt4.QtGui import QStyle
from twisted.internet import defer
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso import USER_CANCELLATION, USER_SUCCESS
from ubuntu_kylin_sso.qt import ssl_dialog
from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.utils.ui import (
    CANCEL_BUTTON,
    SSL_APPNAME_HELP,
    SSL_DETAILS_HELP,
    SSL_DESCRIPTION,
    SSL_DOMAIN_HELP,
    SSL_HEADER,
    SSL_EXPLANATION,
    SSL_FIRST_REASON,
    SSL_SECOND_REASON,
    SSL_THIRD_REASON,
    SSL_CERT_DETAILS,
    SSL_NOT_SURE,
    SSL_REMEMBER_DECISION,
    SSL_HELP_BUTTON,
    SSL_CONNECT_BUTTON,
)

# pylint: disable=W0212, C0103,


class SSLDialogTestCase(TestCase):
    """Test the ssl dialog."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the tests."""
        yield super(SSLDialogTestCase, self).setUp()
        self.domain = 'test-domain'
        self.details = 'SSL details'
        self.appname = 'test'

        self.dialog = ssl_dialog.SSLDialog(self.appname,
                                           domain=self.domain,
                                           details=self.details)
        self.return_code = None

        def fake_done(code):
            """Fake done for the dialog."""
            self.return_code = code

        self.patch(self.dialog, 'done', fake_done)

    def test_init_none_domain(self):
        """Test the init method when the domain is none."""
        dialog = ssl_dialog.SSLDialog(self.appname, domain=None,
                                      details=self.details)
        self.assertEqual(self.details, dialog.details)
        self.assertEqual('', dialog.domain)

    def test_init_none_details(self):
        """Test the init method when the details are none."""
        dialog = ssl_dialog.SSLDialog(self.appname, domain=self.domain,
                                      details=None)
        self.assertEqual('', dialog.details)
        self.assertEqual(self.domain, dialog.domain)

    def test_set_labels(self):
        """Test that the labels contain the correct info."""
        self.assertEqual(SSL_HEADER,
                         compat.text_type(self.dialog.ui.title_label.text()))
        explanation = SSL_EXPLANATION % dict(domain=self.domain)
        intro = ssl_dialog.REASONS_TEMPLATE % dict(explanation=explanation,
                                              first_reason=SSL_FIRST_REASON,
                                              second_reason=SSL_SECOND_REASON,
                                              third_reason=SSL_THIRD_REASON)
        self.assertEqual(intro,
                         compat.text_type(self.dialog.ui.intro_label.text()))
        self.assertEqual(SSL_NOT_SURE % dict(app_name=self.appname),
                compat.text_type(self.dialog.ui.not_sure_label.text()))
        self.assertEqual(SSL_REMEMBER_DECISION,
                compat.text_type(self.dialog.ui.remember_checkbox.text()))

    def test_on_cancel_clicked(self):
        """Test the cancelation action."""
        self.dialog._on_cancel_clicked()
        self.assertEqual(USER_CANCELLATION, self.return_code)

    def test_on_connect_clicked(self):
        """Test the connect action."""
        self.dialog._on_connect_clicked()
        self.assertEqual(USER_SUCCESS, self.return_code)

    def test_set_buttons(self):
        """Test that the buttons are correctly set up."""

        # assert that the buttons have been correctly connected
        called = []

        def fake_on_connect_clicked(event):
            """Fake the on connect clicked method."""
            called.append('_on_connect_clicked')

        self.patch(ssl_dialog.SSLDialog, '_on_connect_clicked',
                                          fake_on_connect_clicked)

        def fake_on_cancel_clicked(event):
            """Fake the on cancle clicked."""
            called.append('_on_cancel_clicked')

        self.patch(ssl_dialog.SSLDialog, '_on_cancel_clicked',
                                          fake_on_cancel_clicked)

        dialog = ssl_dialog.SSLDialog(self.appname, domain=None,
                                      details=self.details)

        dialog.ui.connect_button.clicked.emit(True)
        self.assertIn('_on_connect_clicked', called)

        dialog.ui.cancel_button.clicked.emit(True)
        self.assertIn('_on_cancel_clicked', called)

        self.assertEqual(CANCEL_BUTTON,
                compat.text_type(dialog.ui.cancel_button.text()))
        self.assertEqual(SSL_CONNECT_BUTTON,
                compat.text_type(dialog.ui.connect_button.text()))
        self.assertEqual(SSL_HELP_BUTTON,
                compat.text_type(dialog.ui.help_button.text()))

    def test_set_expander(self):
        """Test that the expander is correctly set."""
        self.assertEqual(SSL_CERT_DETAILS,
            compat.text_type(self.dialog.expander.text()))
        self.assertNotEqual(None, self.dialog.expander.content)
        self.assertEqual(2, self.dialog.ui.expander_layout.indexOf(
                                                        self.dialog.expander))

    def test_set_icon(self):
        """Test that the icon is correctly set."""
        self.assertNotEqual(None, self.dialog.ui.logo_label.pixmap())
        icon = self.dialog.style().standardIcon(QStyle.SP_MessageBoxWarning)
        self.assertEqual(icon.pixmap(48, 48).toImage(),
                         self.dialog.ui.logo_label.pixmap().toImage())
        self.assertEqual('', self.dialog.ui.logo_label.text())


class FakeArgumentParser(object):
    """Fake args parse."""

    def __init__(self):
        """Create an instance."""
        self.called = []

    def __call__(self, description=''):
        """Instance callable."""
        self.called.append(('__init__', description))
        return self

    # pylint: disable=W0622
    def add_argument(self, name, required=False, help=''):
        """Add an argument."""
        self.called.append(('add_argument', name, required, help))
    # pylint: enable=W0622

    def parse_args(self):
        """Parse the args."""
        self.called.append(('parse_args',))
        return self


class ParseArgTestCase(TestCase):
    """Test the parse of the args."""

    def test_parse_args(self):
        """Set tests."""
        argparse = FakeArgumentParser()
        self.patch(ssl_dialog.argparse, 'ArgumentParser', argparse)
        args = ssl_dialog.parse_args()
        self.assertEqual(argparse, args)
        self.assertIn(('__init__', SSL_DESCRIPTION), argparse.called)
        self.assertIn(('add_argument', '--domain', True, SSL_DOMAIN_HELP),
                                                         argparse.called)
        self.assertIn(('add_argument', '--details', True, SSL_DETAILS_HELP),
                                                          argparse.called)
        self.assertIn(('add_argument', '--appname', True, SSL_APPNAME_HELP),
                                                          argparse.called)
        self.assertIn(('parse_args',), argparse.called)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the Setup Account page Qt UI."""

from __future__ import unicode_literals

from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.qt import current_user_sign_in_page as gui
from ubuntu_kylin_sso.qt.tests import (
    PageBaseTestCase,
    FakePageUiStyle,
    FakeWizardButtonStyle,
)
from ubuntu_kylin_sso.tests import EMAIL


# pylint: disable=W0212

class CurrentUserSignInTestCase(PageBaseTestCase):
    """Test the SetupAccountPage code."""

    ui_class = gui.CurrentUserSignInPage
    ui_signals = ('userLoggedIn', 'passwordForgotten', 'userNotValidated',
        'createAccount')
    ui_backend_signals = ('LoggedIn', 'LoginError', 'UserNotValidated')

    def test_initialize_page(self):
        """Test the initialization method."""
        wizard = FakeWizardButtonStyle()
        button = FakePageUiStyle()
        self.patch(self.ui, "wizard", lambda: wizard)
        self.patch(self.ui.ui, "sign_in_button", button)
        self.patch(self.ui, "setButtonText", wizard.setButtonText)
        self.ui.initializePage()
        self.assertEqual(wizard.buttons_text[QtGui.QWizard.CancelButton],
            "Cancel")
        self.assertIn(('setButtonLayout',
            (([],), {})),
            wizard.called)
        self.assertTrue(button.properties['default'])
        self.assertFalse(button.isEnabled())

    def test_unicode_in_forgotten_password_link(self):
        """Ensure that this label supports unicode."""
        forgot_fr = "J'ai oublié mon mot de passe"
        self.patch(gui, "FORGOTTEN_PASSWORD_BUTTON", forgot_fr)
        forgotten_text = gui.LINK_STYLE.format(link_url='#',
                            link_text=forgot_fr)
        self.ui._set_translated_strings()
        self.assertEqual(
                compat.text_type(self.ui.ui.forgot_password_label.text()),
                forgotten_text)

    def test_set_translated_strings(self):
        """Test the translated string method."""
        expected = gui.LOGIN_TITLE.format(app_name=self.app_name)
        self.assertTrue(self.app_name in expected)
        self.assert_title_correct(expected)
        expected = gui.LOGIN_SUBTITLE % dict(app_name=self.app_name)
        self.assertTrue(self.app_name in expected)
        self.assert_subtitle_correct(expected)
        self.assertEqual(compat.text_type(self.ui.ui.email_label.text()),
            gui.EMAIL_LABEL)
        link_text = gui.CREATE_ACCOUNT_LABEL.format(app_name=self.app_name)
        expected = gui.LINK_STYLE.format(link_url='#', link_text=link_text)
        self.assertTrue(self.app_name in expected)
        self.assertEqual(compat.text_type(
            self.ui.ui.create_account_label.text()), expected)
        self.assertEqual(compat.text_type(self.ui.ui.password_label.text()),
            gui.LOGIN_PASSWORD_LABEL)
        text = gui.LINK_STYLE.format(link_url='#',
                                     link_text=gui.FORGOTTEN_PASSWORD_BUTTON)
        self.assertEqual(
                compat.text_type(self.ui.ui.forgot_password_label.text()),
                text)
        self.assertEqual(
                compat.text_type(self.ui.ui.sign_in_button.text()),
                gui.SIGN_IN_BUTTON)

    def test_connect_ui(self):
        """Test the connect ui method."""
        self.assertEqual(self.ui.ui.forgot_password_label.receivers(
            QtCore.SIGNAL("linkActivated(const QString&)")), 1)
        self.assertEqual(self.ui.ui.email_edit.receivers(
            QtCore.SIGNAL("textChanged(const QString&)")), 1)
        self.assertEqual(self.ui.ui.password_edit.receivers(
            QtCore.SIGNAL("textChanged(const QString&)")), 1)
        self.assertEqual(self.ui.ui.sign_in_button.receivers(
            QtCore.SIGNAL("clicked()")), 1)
        self.assertEqual(self.ui.ui.create_account_label.receivers(
            QtCore.SIGNAL("linkActivated(const QString&)")), 1)

    def test_validate_not_valid(self):
        """Test validate method."""
        button = FakePageUiStyle()
        self.patch(self.ui.ui, "sign_in_button", button)
        self.ui.ui.email_edit.setText('not-valid-email')
        self.ui.ui.password_edit.setText('123456')
        self.ui._validate()
        self.assertFalse(button.isEnabled())

    def test_validate_valid(self):
        """Test validate method."""
        button = FakePageUiStyle()
        self.patch(self.ui.ui, "sign_in_button", button)
        self.ui.ui.email_edit.setText('valid@email')
        self.ui.ui.password_edit.setText('123456')
        self.ui._validate()
        self.assertTrue(button.isEnabled())

    def test_login(self):
        """Test the login method."""
        self.patch(self.ui, "hide_error", self._set_called)
        email = 'valid@email'
        password = '123456'
        self.ui.ui.email_edit.setText(email)
        self.ui.ui.password_edit.setText(password)
        self.ui.login()
        self.assertEqual(self._overlay_show_counter, 1)
        self.assertFalse(self.ui.isEnabled())
        self.assert_backend_called('login',
            self.app_name, email, password)
        expected = ((), {})
        self.assertEqual(expected, self._called)

    def test_on_login_error(self):
        """Test the on_login_error method."""
        self.patch(self.ui, "show_error", self._set_called)
        error = {'errtype': 'UserNotValidated'}

        self.ui.on_login_error(self.app_name, error)

        self.assertEqual(self._overlay_hide_counter, 0)
        self.assertTrue(self.ui.isEnabled())
        expected = ((self.ui, self.app_name, ''), {})
        self.assertTrue(expected, self._called)

    def test_on_logged_in(self):
        """Test the on_login_in method."""
        self.ui.ui.email_edit.setText(EMAIL)
        self.assert_signal_emitted(self.ui.userLoggedIn, (EMAIL,),
            self.ui.on_logged_in, self.app_name, EMAIL)
        self.assertTrue(self.ui.isEnabled())

    def test_on_user_not_validated(self):
        """Test the navigation flow on user not validated."""
        self.ui.ui.email_edit.setText(EMAIL)
        self.assert_signal_emitted(self.ui.userNotValidated, (EMAIL,),
            self.ui.on_user_not_validated, self.app_name, EMAIL)

    def test_on_forgotten_password(self):
        """Test the on_forgotten_password method."""
        self.assert_signal_emitted(self.ui.passwordForgotten, (),
            self.ui.on_forgotten_password)

    def test_on_forgotten_password_link_clicked(self):
        """Forgotten passsword clicked emits passwordForgotten signal."""
        self.assert_signal_emitted(self.ui.passwordForgotten, (),
            self.ui.ui.forgot_password_label.linkActivated.emit, "link")

    def test_on_create_account(self):
        """Test the on_create_account method."""
        self.assert_signal_emitted(self.ui.createAccount, (),
            self.ui.on_create_account)

    def test_on_create_account_link_clicked(self):
        """Create account clicked emits createAccount signal."""
        self.assert_signal_emitted(self.ui.createAccount, (),
            self.ui.ui.create_account_label.linkActivated.emit, "link")



#!/usr/bin/python
# setup.py - Build system for Ubuntu SSO Client package
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Setup.py: build, distribute, clean."""

# pylint: disable=W0404, W0511

import distutils
import os
import sys

try:
    import DistUtilsExtra.auto
    from DistUtilsExtra.command import build_extra
except ImportError:
    print >> sys.stderr, 'To build this program you need '\
                         'https://launchpad.net/python-distutils-extra'
    sys.exit(1)
assert DistUtilsExtra.auto.__version__ >= '2.18', \
    'needs DistUtilsExtra.auto >= 2.18'

from distutils import log

from ubuntu_kylin_sso.utils import get_cert_dir

PROJECT_NAME = 'ubuntu-kylin-sso-client'
VERSION = '99.12'

POT_FILE = 'po/ubuntu-kylin-sso-client.pot'
SERVICE_FILE = 'data/com.ubuntukylin.sso.service'
CONSTANTS = 'ubuntu_kylin_sso/constants.py'

CLEANFILES = [SERVICE_FILE, POT_FILE, CONSTANTS, 'MANIFEST']
QT_UI_DIR = os.path.join('ubuntu_kylin_sso', 'qt', 'ui')


def replace_version(*args, **kwargs):
    """Replace the @VERSION@ in the constants file with the actual version."""
    with open(CONSTANTS + '.in') as in_file:
        content = in_file.read()
        with open(CONSTANTS, 'w') as out_file:
            content = content.replace('@VERSION@', VERSION)
            content = content.replace('@PROJECT_NAME@', PROJECT_NAME)
            out_file.write(content)


def replace_prefix(prefix):
    """Replace every '@prefix@' with prefix within 'filename' content."""
    with open(SERVICE_FILE + '.in') as in_file:
        content = in_file.read()
        with open(SERVICE_FILE, 'w') as out_file:
            out_file.write(content.replace('@prefix@', prefix))


class SSOInstall(DistUtilsExtra.auto.install_auto):
    """Class to install proper files."""

    def run(self):
        """Do the install.

        Read from *.service.in and generate .service files by replacing
        @prefix@ by self.prefix.

        """
        prefix = self.install_data.replace(
            self.root if self.root is not None else '', '')
        replace_prefix(prefix)
        DistUtilsExtra.auto.install_auto.run(self)
        # Replace the CONSTANTS prefix here, so that we can do it directly in
        # the installed copy, rather than the lcoal copy. This allows us to
        # have a semi-generated version for use in tests, and a full version
        # for use in installed systems.
        with open(CONSTANTS) as in_file:
            content = in_file.read()
            with open(os.path.join(self.install_purelib,
                                   PROJECT_NAME,
                                   CONSTANTS), 'w') as out_file:
                out_file.write(content.replace('@prefix@', prefix))


class SSOBuild(build_extra.build_extra):
    """Build PyQt (.ui) files and resources."""

    description = "build PyQt GUIs (.ui) and resources (.qrc)"

    def compile_ui(self, ui_file, py_file=None):
        """Compile the .ui files to python modules."""
        # Search for pyuic4 in python bin dir, then in the $Path.
        if py_file is None:
            # go from the ui_file in the data folder to the
            # python file in the qt moodule
            py_file = os.path.split(ui_file)[1]
            py_file = os.path.splitext(py_file)[0] + '_ui.py'
            py_file = os.path.join(QT_UI_DIR, py_file)
        # we indeed want to catch Exception, is ugly but we need it
        # pylint: disable=W0703
        try:
            # import the uic compiler from pyqt and generate the .py files
            # something similar could be done with pyside but that is left
            # as an exercise for the reader.
            from PyQt4 import uic
            fp = open(py_file, 'w')
            uic.compileUi(ui_file, fp)
            fp.close()
            log.info('Compiled %s into %s', ui_file, py_file)
        except Exception, e:
            self.warn('Unable to compile user interface %s: %s' % (py_file, e))
            if not os.path.exists(py_file) or not file(py_file).read():
                raise SystemExit(1)
            return
        # pylint: enable=W0703

    def compile_rc(self, qrc_file, py_file=None):
        """Compile the resources that will be included with the project."""
        import PyQt4
        # Search for pyuic4 in python bin dir, then in the $Path.
        if py_file is None:
            py_file = os.path.split(qrc_file)[1]
            py_file = os.path.splitext(py_file)[0] + '_rc.py'
            py_file = os.path.join(QT_UI_DIR, py_file)
        path = os.getenv('PATH')
        os.putenv('PATH', path + os.path.pathsep + os.path.join(
                  os.path.dirname(PyQt4.__file__), 'bin'))
        if os.system('pyrcc4 -no-compress "%s" -o "%s"' %
                     (qrc_file, py_file)) > 0:
            self.warn('Unable to generate python module {py_file}'
                      ' for resource file {qrc_file}'.format(
                          py_file=py_file, qrc_file=qrc_file))
            if not os.path.exists(py_file) or not file(py_file).read():
                raise SystemExit(1)
        else:
            log.info('compiled %s into %s' % (qrc_file, py_file))
        os.putenv('PATH', path)

    def _generate_qrc(self, qrc_file, srcfiles, prefix):
        """Generate the qrc file for the given src files."""
        basedir = os.path.dirname(qrc_file)
        f = open(qrc_file, 'w')
        try:
            f.write('<!DOCTYPE RCC><RCC version="1.0">\n')
            import cgi
            f.write('  <qresource prefix="%s">\n' % cgi.escape(prefix))
            for e in srcfiles:
                relpath = e[len(basedir) + 1:]
                f.write('    <file>%s</file>\n'
                        % cgi.escape(relpath.replace(os.path.sep, '/')))
            f.write('  </qresource>\n')
            f.write('</RCC>\n')
        finally:
            f.close()

    def build_rc(self, py_file, basedir, prefix='/'):
        """Generate compiled resource including any files under basedir"""
        # For details, see http://doc.qt.nokia.com/latest/resources.html
        qrc_file = os.path.join(basedir, '%s.qrc' % os.path.basename(basedir))
        srcfiles = [os.path.join(root, e)
                    for root, _dirs, files in os.walk(basedir) for e in files]
        # NOTE: Here we cannot detect deleted files. In such cases, we need
        # to remove .qrc manually.
        try:
            self._generate_qrc(qrc_file, srcfiles, prefix)
            self.compile_rc(qrc_file, py_file)
        finally:
            os.unlink(qrc_file)

    def run(self):
        """Execute the command."""
        replace_version()
        basepath = os.path.join('data', 'qt')
        # TODO: build the resource files so that we can include them
        # self.build_rc(os.path.join(basepath, 'icons_rc.py'),
        #              os.path.join(os.path.dirname(__file__), 'icons'),
        #              '/icons')
        for dirpath, _, filenames in os.walk(basepath):
            for filename in filenames:
                if filename.endswith('.ui'):
                    self.compile_ui(os.path.join(dirpath, filename))
                elif filename.endswith('.qrc'):
                    self.compile_rc(os.path.join(dirpath, filename))

        build_extra.build_extra.run(self)


class SSOClean(DistUtilsExtra.auto.clean_build_tree):
    """Class to clean up after the build."""

    def run(self):
        """Clean up the built files."""
        for built_file in CLEANFILES:
            if os.path.exists(built_file):
                os.unlink(built_file)

        for dirpath, _, filenames in os.walk(os.path.join(QT_UI_DIR)):
            for current_file in filenames:
                if current_file.endswith('_ui.py') or\
                                current_file.endswith('_rc.py'):
                    os.unlink(os.path.join(dirpath, current_file))

        DistUtilsExtra.auto.clean_build_tree.run(self)


def set_py2exe_paths():
    """Set the path so that py2exe finds the required modules."""
    # Pylint does not understand same spaced imports
    # pylint: disable=F0401
    import win32com
    # pylint: enable=F0401
    try:
        # pylint: disable=F0401
        import py2exe.mf as modulefinder
        # pylint: enable=F0401
    except ImportError:
        import modulefinder

    # py2exe 0.6.4 introduced a replacement modulefinder.
    # This means we have to add package paths there,
    # not to the built-in one.  If this new modulefinder gets
    # integrated into Python, then we might be able to revert
    # this some day. If this doesn't work, try import modulefinder
    for package_path in win32com.__path__[1:]:
        modulefinder.AddPackagePath("win32com", package_path)
    for extra_mod in ["win32com.server", "win32com.client"]:
        __import__(extra_mod)
        module = sys.modules[extra_mod]
        for module_path in module.__path__[1:]:
            modulefinder.AddPackagePath(extra_mod, module_path)


# pylint: disable=C0103

cmdclass = {
    'install': SSOInstall,
    'build': SSOBuild,
    'clean': SSOClean,
}


sso_executables = [
    'bin/ubuntu-kylin-sso-login',
    'bin/ubuntu-kylin-sso-login-qt',
    'bin/ubuntu-kylin-sso-proxy-creds-qt',
    'bin/ubuntu-kylin-sso-ssl-certificate-qt',
]


class dummy_build_i18n(distutils.cmd.Command):

    """Dummy for windows."""

    def initialize_options(self, *args):
        """Dummy."""

    def finalize_options(self, *args):
        """Dummy."""

    def run(self, *args):
        """Dummy."""


if sys.platform == 'win32':
    cmdclass['build_i18n'] = dummy_build_i18n


# data_files = [(get_cert_dir(),
#              ['data/UbuntuOne-Go_Daddy_CA.pem',
#               'data/UbuntuOne-ValiCert_Class_2_VA.pem',
#               'data/UbuntuOne-Go_Daddy_Class_2_CA.pem'])]
data_files = [(get_cert_dir(),
             ['data/UbuntuKylin_CA.pem'])]

if sys.platform == 'win32':
    set_py2exe_paths()
    # for PyQt, see http://www.py2exe.org/index.cgi/Py2exeAndPyQt
    _includes = ['sip', 'email', 'ubuntu_kylin_sso.qt.gui',
                 'ubuntu_kylin_sso.qt.controllers', 'PyQt4.QtNetwork', 'PIL']
    extra = {
        'options': {
            'py2exe': {
                'bundle_files': 1,
                'skip_archive': 0,
                'includes': _includes,
                'optimize': 1,
                'dll_excludes': ["mswsock.dll", "powrprof.dll"],
            },
        },
        # add the console script so that py2exe compiles it
        'console': sso_executables,
        'zipfile': None,
    }
else:
    data_files.extend([
        ('lib/ubuntu-kylin-sso-client', sso_executables),
        ('share/dbus-1/services', ['data/com.ubuntukylin.sso.service']),
        ('share/app-install/icons', ['data/youker-id.png']),
    ])
    extra = {}

DistUtilsExtra.auto.setup(
    name=PROJECT_NAME,
    version=VERSION,
    license='GPL v3',
    author='Ubuntu Kylin Team',
    author_email='ubuntukylin-members@list.launchpad.net',
    description='Ubuntu Kylin Single Sign-On client',
    long_description=('Desktop service to allow applications to sign in'
                      'to Ubuntu Kylin services via SSO'),
    url='https://launchpad.net/ubuntu-kylin-sso-client',
    extra_path=PROJECT_NAME,
    data_files=data_files,
    packages=[
        'ubuntu_kylin_sso',
        'ubuntu_kylin_sso.tests',
        'ubuntu_kylin_sso.keyring',
        'ubuntu_kylin_sso.keyring.tests',
        'ubuntu_kylin_sso.main',
        'ubuntu_kylin_sso.main.tests',
        'ubuntu_kylin_sso.networkstate',
        'ubuntu_kylin_sso.networkstate.tests',
        'ubuntu_kylin_sso.qt',
        'ubuntu_kylin_sso.qt.main',
        'ubuntu_kylin_sso.qt.ui',
        'ubuntu_kylin_sso.utils',
        'ubuntu_kylin_sso.utils.tests',
        'ubuntu_kylin_sso.utils.runner',
        'ubuntu_kylin_sso.utils.runner.tests',
        'ubuntu_kylin_sso.utils.webclient',
        'ubuntu_kylin_sso.utils.webclient.tests',
    ],
    cmdclass=cmdclass,
    **extra)



# -*- coding: utf-8 -*-
#
# Copyright 2009-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the main SSO client code."""

from __future__ import unicode_literals

import logging

from functools import partial

from twisted.internet import defer
from ubuntuone.devtools.handlers import MementoHandler

from ubuntu_kylin_sso import main
from ubuntu_kylin_sso.main import (
    CredentialsManagement,
    except_to_errdict,
    SSOLogin,
    TIMEOUT_INTERVAL,
    UbuntuSSOService,
)
from ubuntu_kylin_sso.main import (
    HELP_TEXT_KEY,
    PING_URL_KEY,
    POLICY_URL_KEY,
    TC_URL_KEY,
    UI_EXECUTABLE_KEY,
    WINDOW_ID_KEY,
)
from ubuntu_kylin_sso.main.tests import FakedCredentials
from ubuntu_kylin_sso.tests import (
    APP_NAME,
    CAPTCHA_ID,
    CAPTCHA_SOLUTION,
    EMAIL,
    EMAIL_TOKEN,
    HELP_TEXT,
    NAME,
    PASSWORD,
    PING_URL,
    POLICY_URL,
    Recorder,
    TC_URL,
    TestCase,
    TOKEN,
    WINDOW_ID,
)


# Access to a protected member 'yyy' of a client class
# pylint: disable=W0212


class SampleException(Exception):
    """A sample exception."""


class FakedProxy(Recorder):
    """A faked multiplatform proxy."""


class FakedAccount(Recorder):
    """A faked Account processor."""


class FakedCredentialsFactory(Recorder):
    """A very dummy Credentials object."""

    credentials = FakedCredentials()

    # pylint: disable=C0103

    def Credentials(self, *a, **kw):
        """Return always the same Credentials instance."""
        return self.credentials

    # pylint: enable=C0103


class TestExceptToErrdictException(Exception):
    """A dummy exception for the following testcase."""


class ExceptToErrdictTestCase(TestCase):
    """Tests for the except_to_errdict function."""

    def test_first_arg_is_dict(self):
        """If the first arg is a dict, use it as the base dict."""
        sample_dict = {
            "errorcode1": "error message 1",
            "errorcode2": "error message 2",
            "errorcode3": "error message 3",
        }
        e = TestExceptToErrdictException(sample_dict)
        result = except_to_errdict(e)

        self.assertEqual(result["errtype"], e.__class__.__name__)
        for k in sample_dict.keys():
            self.assertIn(k, result)
            self.assertEqual(result[k], sample_dict[k])

    def test_first_arg_is_str(self):
        """If the first arg is a str, use it as the message."""
        sample_string = "a sample string"
        e = TestExceptToErrdictException(sample_string)
        result = except_to_errdict(e)
        self.assertEqual(result["errtype"], e.__class__.__name__)
        self.assertEqual(result["message"], sample_string)

    def test_first_arg_is_exception(self):
        """If the first arg is an Exception, use it's message attribute'."""

        class MyException(Exception):
            """Custom Exception."""

        message = 'My custom error for ♥ Ubuntu'
        my_exc = MyException(message)
        result = except_to_errdict(my_exc)
        self.assertEqual(result["errtype"], my_exc.__class__.__name__)
        self.assertEqual(result["message"], message)

    def test_first_arg_is_unicode(self):
        """If the first arg is a unicode, use it as the message."""
        sample_string = "a sample string"
        e = TestExceptToErrdictException(sample_string)
        result = except_to_errdict(e)
        self.assertEqual(result["errtype"], e.__class__.__name__)
        self.assertEqual(result["message"], sample_string)

    def test_no_args_at_all(self):
        """If there are no args, use the class docstring."""
        e = TestExceptToErrdictException()
        result = except_to_errdict(e)
        self.assertEqual(result["errtype"], e.__class__.__name__)
        self.assertEqual(result["message"], e.__class__.__doc__)

    def test_some_other_thing_as_first_arg(self):
        """If first arg is not basestring nor dict, then repr all args."""
        sample_args = (None, "unicode2\ufffd", "errorcode3")
        e = TestExceptToErrdictException(*sample_args)
        result = except_to_errdict(e)
        self.assertEqual(result["errtype"], e.__class__.__name__)


class BaseTestCase(TestCase):
    """Base test case."""

    timeout = 2

    @defer.inlineCallbacks
    def setUp(self):
        yield super(BaseTestCase, self).setUp()
        self.proxy = FakedProxy()
        self.deferred = defer.Deferred()

    def assert_recorder_called(self, fake, method, *args, **kwargs):
        """Check that 'fake.method(*args, **kwargs)' was called."""
        self.assertEqual(fake._called[method], [(args, kwargs)])

    def callback_deferred(self, *a):
        """Callback self.deferred with a as result."""
        self.deferred.callback(a)

    def errback_deferred(self, signal_name, *a):
        """Errback self.deferred with an AssertionError as error."""
        msg = 'Received unexpected signal %r with params %r.'
        self.deferred.errback(AssertionError(msg % (signal_name, a)))


class SSOLoginBaseTestCase(BaseTestCase):
    """Base test case for testing the SSOLogin class."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(SSOLoginBaseTestCase, self).setUp()

        def ksc(keyring, k, val):
            """Assert over token and app_name."""
            self.assertEqual(k, APP_NAME)
            self.assertEqual(val, TOKEN)
            self.keyring_was_set = True
            return defer.succeed(None)

        self.patch(main.Keyring, "set_credentials", ksc)
        self.patch(main, 'Account', FakedAccount)
        self.keyring_was_set = False

        self.obj = SSOLogin(self.proxy)

    def test_creation(self):
        """Test that the object creation is successful."""
        self.assertIsInstance(self.obj.processor, main.Account)
        self.assertTrue(self.obj.proxy is self.proxy)


class SSOLoginTestCase(SSOLoginBaseTestCase):
    """Test the SSOLogin class."""

    @defer.inlineCallbacks
    def test_generate_captcha(self):
        """Test that the captcha method works ok."""
        filename = "sample filename"
        expected_result = "expected result"

        self.patch(self.obj, "CaptchaGenerated", self.callback_deferred)
        self.patch(self.obj, "CaptchaGenerationError",
                   partial(self.errback_deferred, "CaptchaGenerationError"))

        self.obj.processor._next_result = defer.succeed(expected_result)
        self.obj.generate_captcha(APP_NAME, filename)

        app_name, result = yield self.deferred
        self.assertEqual(result, expected_result)
        self.assertEqual(app_name, APP_NAME)
        self.assert_recorder_called(self.obj.processor, 'generate_captcha',
            filename)

    @defer.inlineCallbacks
    def test_register_user(self):
        """Test that the register_user method works ok."""
        expected_result = "expected result"

        self.patch(self.obj, "UserRegistered", self.callback_deferred)
        self.patch(self.obj, "UserRegistrationError",
                   partial(self.errback_deferred, "UserRegistrationError"))

        self.obj.processor._next_result = defer.succeed(expected_result)
        self.obj.register_user(APP_NAME, EMAIL, PASSWORD, NAME, CAPTCHA_ID,
                               CAPTCHA_SOLUTION)

        app_name, result = yield self.deferred
        self.assertEqual(result, expected_result)
        self.assertEqual(app_name, APP_NAME)
        self.assert_recorder_called(self.obj.processor, 'register_user',
            EMAIL, PASSWORD, NAME, CAPTCHA_ID, CAPTCHA_SOLUTION)

    @defer.inlineCallbacks
    def test_login(self):
        """Test that the login method works ok when used is validated."""
        self.patch(self.obj, "LoggedIn", self.callback_deferred)
        self.patch(self.obj, "LoginError",
                   partial(self.errback_deferred, "LoginError"))
        self.patch(self.obj, "UserNotValidated",
                   partial(self.errback_deferred, "UserNotValidated"))
        self.patch(self.obj.processor, 'is_validated', lambda _: True)

        self.obj.processor._next_result = TOKEN
        self.obj.login(APP_NAME, EMAIL, PASSWORD)

        app_name, result = yield self.deferred
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(self.keyring_was_set, "The keyring should be set")
        self.assert_recorder_called(self.obj.processor, 'login',
            EMAIL, PASSWORD, main.get_token_name(APP_NAME))
        self.assert_recorder_called(self.obj.processor, 'is_validated',
            TOKEN)

    @defer.inlineCallbacks
    def test_login_is_not_validated(self):
        """Test that the login method works ok when user is not validated."""
        self.patch(self.obj, "LoggedIn",
                   partial(self.errback_deferred, "LoggedIn"))
        self.patch(self.obj, "LoginError",
                   partial(self.errback_deferred, "LoginError"))
        self.patch(self.obj, "UserNotValidated", self.callback_deferred)
        self.patch(self.obj.processor, 'is_validated', lambda _: False)

        self.obj.processor._next_result = TOKEN
        self.obj.login(APP_NAME, EMAIL, PASSWORD)

        app_name, result = yield self.deferred
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assertFalse(self.keyring_was_set, "The keyring should not be set")
        self.assert_recorder_called(self.obj.processor, 'login',
            EMAIL, PASSWORD, main.get_token_name(APP_NAME))
        self.assert_recorder_called(self.obj.processor, 'is_validated',
            TOKEN)

    @defer.inlineCallbacks
    def test_validate_email(self):
        """Test that the validate_email method works ok."""
        self.patch(self.obj, "EmailValidated", self.callback_deferred)
        self.patch(self.obj, "EmailValidationError",
                   partial(self.errback_deferred, "EmailValidationError"))

        self.obj.processor._next_result = TOKEN
        self.obj.validate_email(APP_NAME, EMAIL, PASSWORD, EMAIL_TOKEN)

        app_name, result = yield self.deferred
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(self.keyring_was_set, "The keyring should be set")
        self.assert_recorder_called(self.obj.processor, 'validate_email',
            EMAIL, PASSWORD, EMAIL_TOKEN, main.get_token_name(APP_NAME))

    @defer.inlineCallbacks
    def test_request_password_reset_token(self):
        """Test that the request_password_reset_token method works ok."""
        self.patch(self.obj, "PasswordResetTokenSent", self.callback_deferred)
        self.patch(self.obj, "PasswordResetError",
                   partial(self.errback_deferred, "PasswordResetError"))

        self.obj.processor._next_result = EMAIL
        self.obj.request_password_reset_token(APP_NAME, EMAIL)

        app_name, result = yield self.deferred
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assert_recorder_called(self.obj.processor,
            'request_password_reset_token', EMAIL)

    @defer.inlineCallbacks
    def test_set_new_password(self):
        """Test that the set_new_password method works ok."""
        self.patch(self.obj, "PasswordChanged", self.callback_deferred)
        self.patch(self.obj, "PasswordChangeError",
                   partial(self.errback_deferred, "PasswordChangeError"))

        self.obj.processor._next_result = EMAIL
        self.obj.set_new_password(APP_NAME, EMAIL, EMAIL_TOKEN, PASSWORD)

        app_name, result = yield self.deferred
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assert_recorder_called(self.obj.processor,
            'set_new_password', EMAIL, EMAIL_TOKEN, PASSWORD)


class SSOLoginWithErrorTestCase(SSOLoginTestCase):
    """Test the SSOLogin class."""

    def sample_fail(self, *args):
        """Fail predictably."""
        raise SampleException("Sample message")

    @defer.inlineCallbacks
    def test_generate_captcha(self):
        """Test that the captcha method fails as expected."""
        filename = "sample filename"

        self.patch(self.obj, "CaptchaGenerated",
                   partial(self.errback_deferred, "CaptchaGenerated"))
        self.patch(self.obj, "CaptchaGenerationError", self.callback_deferred)
        self.patch(self.obj.processor, "generate_captcha", self.sample_fail)
        self.obj.generate_captcha(APP_NAME, filename)

        app_name, error = yield self.deferred
        self.assertTrue(isinstance(error, SampleException))
        self.assertEqual(app_name, APP_NAME)

    @defer.inlineCallbacks
    def test_register_user(self):
        """Test that the register_user method fails as expected."""
        self.patch(self.obj, "UserRegistered",
                   partial(self.errback_deferred, "UserRegistered"))
        self.patch(self.obj, "UserRegistrationError", self.callback_deferred)
        self.patch(self.obj.processor, "register_user", self.sample_fail)
        self.obj.register_user(APP_NAME, EMAIL, PASSWORD, NAME, CAPTCHA_ID,
                             CAPTCHA_SOLUTION)

        app_name, error = yield self.deferred
        self.assertTrue(isinstance(error, SampleException))
        self.assertEqual(app_name, APP_NAME)

    @defer.inlineCallbacks
    def test_login(self):
        """Test that the login method works ok."""

        def fake_gtn(*args):
            """A fake get_token_name that fails."""
            raise SampleException()

        self.patch(main, "get_token_name", fake_gtn)
        self.patch(self.obj, "LoggedIn",
                   partial(self.errback_deferred, "LoggedIn"))
        self.patch(self.obj, "LoginError", self.callback_deferred)
        self.patch(self.obj, "UserNotValidated",
                   partial(self.errback_deferred, "UserNotValidated"))
        self.obj.login(APP_NAME, EMAIL, PASSWORD)

        app_name, error = yield self.deferred
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(isinstance(error, SampleException))
        self.assertFalse(self.keyring_was_set, "Keyring should not be set")

    @defer.inlineCallbacks
    def test_login_is_not_validated(self):
        """Test that the login method works ok."""

        def fake_validate(*args):
            """A fake get_token_name that fails."""
            raise SampleException()

        self.patch(self.obj, "LoggedIn",
                   partial(self.errback_deferred, "LoggedIn"))
        self.patch(self.obj, "LoginError", self.callback_deferred)
        self.patch(self.obj, "UserNotValidated",
                   partial(self.errback_deferred, "UserNotValidated"))
        self.patch(self.obj.processor, 'is_validated', fake_validate)
        self.obj.login(APP_NAME, EMAIL, PASSWORD)

        app_name, error = yield self.deferred
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(isinstance(error, SampleException))
        self.assertFalse(self.keyring_was_set, "Keyring should not be set")

    @defer.inlineCallbacks
    def test_login_set_credentials(self):
        """The login method fails as expected when set_credentials fails."""

        def fake_set_creds(*args):
            """A fake Keyring.set_credentials that fails."""
            return defer.fail(SampleException())

        self.patch(main.Keyring, "set_credentials", fake_set_creds)
        self.patch(self.obj, "LoggedIn",
                   partial(self.errback_deferred, "LoggedIn"))
        self.patch(self.obj, "LoginError", self.callback_deferred)
        self.patch(self.obj, "UserNotValidated",
                   partial(self.errback_deferred, "UserNotValidated"))
        self.patch(self.obj.processor, "login",
                                        lambda *args: defer.succeed(None))
        self.patch(self.obj.processor, "is_validated",
                                        lambda *args: defer.succeed(True))
        self.obj.login(APP_NAME, EMAIL, PASSWORD)

        app_name, error = yield self.deferred
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(isinstance(error, SampleException))
        self.assertFalse(self.keyring_was_set, "Keyring should not be set")

    @defer.inlineCallbacks
    def test_validate_email(self):
        """Test that the validate_email method fails as expected."""

        def fake_gtn(*args):
            """A fake get_token_name that fails."""
            raise SampleException()

        self.patch(main, "get_token_name", fake_gtn)

        self.patch(self.obj, "EmailValidated",
                   partial(self.errback_deferred, "EmailValidated"))
        self.patch(self.obj, "EmailValidationError", self.callback_deferred)
        self.obj.validate_email(APP_NAME, EMAIL, PASSWORD, EMAIL_TOKEN)

        app_name, error = yield self.deferred
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(isinstance(error, SampleException))
        self.assertFalse(self.keyring_was_set, "Keyring should not be set")

    @defer.inlineCallbacks
    def test_request_password_reset_token(self):
        """Test the request_password_reset_token method fails as expected."""
        self.patch(self.obj, "PasswordResetTokenSent",
                   partial(self.errback_deferred, "PasswordResetTokenSent"))
        self.patch(self.obj, "PasswordResetError", self.callback_deferred)
        self.patch(self.obj.processor, "request_password_reset_token",
                   lambda _: defer.fail(SampleException()))
        self.obj.request_password_reset_token(APP_NAME, EMAIL)

        app_name, error = yield self.deferred
        self.assertTrue(isinstance(error, SampleException))
        self.assertEqual(app_name, APP_NAME)

    @defer.inlineCallbacks
    def test_set_new_password(self):
        """Test that the set_new_password method fails as expected."""
        self.patch(self.obj, "PasswordChanged",
                   partial(self.errback_deferred, "PasswordChanged"))
        self.patch(self.obj, "PasswordChangeError", self.callback_deferred)
        self.patch(self.obj.processor, "set_new_password",
                   lambda *args: defer.fail(SampleException()))
        self.obj.set_new_password(APP_NAME, EMAIL, EMAIL_TOKEN, PASSWORD)

        app_name, error = yield self.deferred
        self.assertTrue(isinstance(error, SampleException))
        self.assertEqual(app_name, APP_NAME)


class PingUrlOnNewToken(SSOLoginBaseTestCase):
    """The test case for the operations that pings a given url."""

    method = 'login'
    args = (APP_NAME, EMAIL, PASSWORD, PING_URL)
    success_signal = 'LoggedIn'
    error_signal = 'LoginError'
    unexpected_signals = ('UserNotValidated',)

    @defer.inlineCallbacks
    def setUp(self):
        yield super(PingUrlOnNewToken, self).setUp()
        self.patch(main.utils, 'ping_url', self.my_ping)
        self.patch(self.obj.processor, 'is_validated', lambda _: True)
        self.obj.processor._next_result = TOKEN
        self._next_ping_result = defer.succeed('foo')

        for i in self.unexpected_signals:
            self.patch(self.obj, i, partial(self.errback_deferred, i))

    def my_ping(self, *a):
        """Fake the ping_url method."""
        self._called = a
        return self._next_ping_result

    @defer.inlineCallbacks
    def test_url_is_pinged(self):
        """The method calls self.success_signal when ping was completed."""
        self.patch(self.obj, self.success_signal, self.callback_deferred)
        self.patch(self.obj, self.error_signal,
                   partial(self.errback_deferred, self.error_signal))

        getattr(self.obj, self.method)(*self.args)

        app_name, result = yield self.deferred

        # url was pinged with the utils.ping_url method
        self.assertEqual(self._called, (PING_URL, EMAIL, TOKEN))
        # token was stored in the keyring
        self.assertTrue(self.keyring_was_set, "The keyring should be set")

        self.assertEqual(app_name, APP_NAME)
        self.assertEqual(result, EMAIL)

    @defer.inlineCallbacks
    def test_ping_url_raises_exception(self):
        """Exception is handled if ping fails."""
        self.patch(self.obj, self.error_signal, self.callback_deferred)
        self.patch(self.obj, self.success_signal,
                   partial(self.errback_deferred, self.success_signal))

        error = 'Blu'
        self._next_ping_result = defer.fail(ValueError(error))
        getattr(self.obj, self.method)(*self.args)

        app_name, error = yield self.deferred

        # token was not stored in the keyring
        self.assertFalse(self.keyring_was_set, "The keyring should not be set")
        self.assertEqual(app_name, APP_NAME)
        self.assertIsInstance(error, ValueError)
        self.assertIn('Blu', error.message)


class PingUrlOnNewTokenEmailValidated(PingUrlOnNewToken):
    """The test case for the operations that pings a given url."""

    method = 'validate_email'
    args = (APP_NAME, EMAIL, PASSWORD, EMAIL_TOKEN, PING_URL)
    success_signal = 'EmailValidated'
    error_signal = 'EmailValidationError'
    unexpected_signals = ()


class CredentialsManagementTestCase(BaseTestCase):
    """Tests for the CredentialsManagement DBus interface."""

    base_args = {
        HELP_TEXT_KEY: HELP_TEXT, PING_URL_KEY: PING_URL,
        POLICY_URL_KEY: POLICY_URL,
        TC_URL_KEY: TC_URL, WINDOW_ID_KEY: WINDOW_ID,
        UI_EXECUTABLE_KEY: 'super-ui',
    }

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementTestCase, self).setUp()

        self.factory = FakedCredentialsFactory()
        self.patch(main, 'Credentials', self.factory.Credentials)
        self.obj = CredentialsManagement(timeout_func=lambda *a: None,
                                         shutdown_func=lambda *a: None,
                                         proxy=self.proxy)
        self.args = {}
        self.cred_args = {}

        self.memento = MementoHandler()
        self.memento.setLevel(logging.DEBUG)
        main.logger.addHandler(self.memento)
        self.addCleanup(main.logger.removeHandler, self.memento)


class CredentialsManagementRefCountingTestCase(CredentialsManagementTestCase):
    """Tests for the CredentialsManagement ref counting."""

    @defer.inlineCallbacks
    def assert_refcounter_increased(self, method_name, signal_name=None):
        """Assert that calling 'method_name' increases the ref counter."""
        if signal_name is not None:
            d = defer.Deferred()
            self.patch(self.obj, signal_name, lambda *a: d.callback(a))
        else:
            d = defer.succeed(None)

        getattr(self.obj, method_name)(APP_NAME, self.args)

        yield d
        self.assertEqual(self.obj.ref_count, 1)

    def assert_refcounter_decreased(self, signal_name, *a, **kw):
        """Assert that calling 'signal_name' decreases the ref counter."""
        self.obj.ref_count = 3

        getattr(self.obj, signal_name)(*a, **kw)

        self.assertEqual(self.obj.ref_count, 2)

    def assert_refcounter_negative(self, signal_name, *a, **kw):
        """Assert that calling 'signal_name' decreases the ref counter.

        If decreased value is negative, assert that a warning log was made.

        """
        self.obj._ref_count = -3  # overwrite internal to force a negative

        getattr(self.obj, signal_name)(*a, **kw)

        self.assertEqual(self.obj.ref_count, 0)
        msg = 'Attempting to decrease ref_count to a negative value (-4).'
        self.assertTrue(self.memento.check_warning(msg))

    def test_ref_counting(self):
        """Ref counting is in place."""
        self.assertEqual(self.obj.ref_count, 0)

    @defer.inlineCallbacks
    def test_find_credentials(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('find_credentials',
                                               'CredentialsFound')

    @defer.inlineCallbacks
    def test_find_credentials_with_success_cb(self):
        """Keep proper track of ongoing requests."""
        d = defer.Deferred()
        counter = []

        def in_the_middle():
            """Store the ref_count value."""
            counter.append(self.obj.ref_count)
            return defer.succeed(TOKEN)

        self.patch(self.factory.credentials, 'find_credentials', in_the_middle)
        self.obj.find_credentials(APP_NAME, self.args,
                                  success_cb=d.callback, error_cb=d.errback)
        yield d
        self.assertEqual(counter, [1])
        self.assertEqual(self.obj.ref_count, 0)

    @defer.inlineCallbacks
    def test_find_credentials_with_error_cb(self):
        """Keep proper track of ongoing requests."""
        d = defer.Deferred()
        counter = []

        def in_the_middle():
            """Store the ref_count value."""
            counter.append(self.obj.ref_count)
            return defer.fail(SampleException('foo'))

        self.patch(self.factory.credentials, 'find_credentials', in_the_middle)
        self.obj.find_credentials(APP_NAME, self.args,
                                  success_cb=d.errback, error_cb=d.callback)

        yield d
        self.assertEqual(counter, [1])
        self.assertEqual(self.obj.ref_count, 0)

    @defer.inlineCallbacks
    def test_clear_credentials(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('clear_credentials',
                                               'CredentialsCleared')

    @defer.inlineCallbacks
    def test_store_credentials(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('store_credentials',
                                               'CredentialsStored')

    @defer.inlineCallbacks
    def test_register(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('register',
                                               'CredentialsFound')

    @defer.inlineCallbacks
    def test_login(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('login',
                                               'CredentialsFound')

    def test_several_requests(self):
        """Requests can be nested."""
        # by patching the CredentialsFound signal, the counter is not decreased
        self.patch(self.obj, 'CredentialsFound', lambda *a: None)

        self.obj.login(APP_NAME, self.args)
        self.obj.register(APP_NAME, self.args)
        self.obj.login(APP_NAME, self.args)
        self.obj.register(APP_NAME, self.args)
        self.obj.register(APP_NAME, self.args)

        self.assertEqual(self.obj.ref_count, 5)

    def test_credentials_found(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsFound', APP_NAME, TOKEN)

    def test_credentials_not_found(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsNotFound', APP_NAME)

    def test_credentials_cleared(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsCleared', APP_NAME)

    def test_credentials_stored(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsStored', APP_NAME)

    def test_credentials_error(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsError', APP_NAME,
                                         SampleException('test'))

    def test_authorization_denied(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('AuthorizationDenied', APP_NAME)

    def test_credentials_found_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsFound', APP_NAME, TOKEN)

    def test_credentials_not_found_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsNotFound', APP_NAME)

    def test_credentials_cleared_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsCleared', APP_NAME)

    def test_credentials_stored_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsStored', APP_NAME)

    def test_credentials_error_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsError', APP_NAME,
                                        SampleException('test'))

    def test_autorization_denied_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('AuthorizationDenied', APP_NAME)

    def test_on_zero_ref_count_shutdown(self):
        """When ref count reaches 0, queue shutdown op."""
        self.patch(self.obj, 'timeout_func', self._set_called)
        self.obj.login(APP_NAME, self.args)

        self.assertEqual(self._called,
                         ((TIMEOUT_INTERVAL, self.obj.shutdown), {}))

    def test_on_non_zero_ref_count_do_not_shutdown(self):
        """If ref count is not 0, do not queue shutdown op."""
        self.patch(self.obj, 'timeout_func', self._set_called)
        # by patching the CredentialsFound signal, the counter is not decreased
        self.patch(self.obj, 'CredentialsFound', lambda *a: None)
        self.obj.login(APP_NAME, self.args)

        self.assertEqual(self._called, False)

    def test_on_non_zero_ref_count_after_zero_do_not_shutdown(self):
        """If the shutdown was queued, do not quit if counter is not zero."""

        def fake_timeout_func(interval, func):
            """Start a new request when the timer is started."""
            # the counter is not decreased
            self.patch(self.obj, 'CredentialsFound', lambda *a: None)
            self.obj.register(APP_NAME, self.args)
            assert self.obj.ref_count > 0
            func()

        self.patch(self.obj, 'timeout_func', fake_timeout_func)
        self.patch(self.obj, 'shutdown_func', self._set_called)

        self.obj.login(APP_NAME, self.args)
        # counter reached 0, timeout_func was called and register was called

        self.assertEqual(self._called, False, 'shutdown_func was not called')

    def test_zero_ref_count_after_zero_do_shutdown(self):
        """If the shutdown was queued, do quit if counter is zero."""

        def fake_timeout_func(interval, func):
            """Start a new request when the timer is started."""
            assert self.obj.ref_count == 0
            func()

        self.patch(self.obj, 'timeout_func', fake_timeout_func)
        self.patch(self.obj, 'shutdown_func', self._set_called)

        self.obj.login(APP_NAME, self.args)
        # counter reached 0, timeout_func was called

        self.assertEqual(self._called, ((), {}), 'shutdown_func was called')


class CredentialsManagementClearTestCase(CredentialsManagementTestCase):
    """Tests for the CredentialsManagement clear method."""

    method = 'clear_credentials'
    success_signal = 'CredentialsCleared'
    success_result = (APP_NAME,)
    others_should_errback = []

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementClearTestCase, self).setUp()
        self.call_method = getattr(self.obj, self.method)

    def test_backend_called(self):
        """The credentials backend is properly called."""
        self.call_method(APP_NAME, self.args)
        self.assert_recorder_called(self.factory, 'Credentials', APP_NAME)

    @defer.inlineCallbacks
    def test_does_not_block(self):
        """Calling 'method' does not block but return thru signals."""
        self.patch(self.obj, self.success_signal, self.callback_deferred)
        self.patch(self.obj, 'CredentialsError',
                   partial(self.errback_deferred, 'CredentialsError'))
        for signal in self.others_should_errback:
            self.patch(self.obj, signal,
                       partial(self.errback_deferred, signal))

        self.call_method(APP_NAME, self.args)

        result = yield self.deferred
        self.assertEqual(result, self.success_result)

    @defer.inlineCallbacks
    def test_handles_error(self):
        """If calling the backend fails, CredentialsError is sent."""
        self.patch(self.obj, self.success_signal,
                   partial(self.errback_deferred, self.success_signal))
        self.patch(self.obj, 'CredentialsError', self.callback_deferred)

        exc = SampleException('baz')
        self.patch(self.factory.credentials, self.method,
                   lambda *a: defer.fail(exc))
        self.call_method(APP_NAME, self.args)

        app_name, error = yield self.deferred
        self.assertEqual(error, exc)
        self.assertEqual(app_name, APP_NAME)


class CredentialsManagementStoreTestCase(CredentialsManagementClearTestCase):
    """Tests for the CredentialsManagement store method."""

    method = 'store_credentials'
    success_signal = 'CredentialsStored'


class CredentialsManagementFindTestCase(CredentialsManagementClearTestCase):
    """Tests for the CredentialsManagement find method."""

    method = 'find_credentials'
    success_signal = 'CredentialsFound'
    success_result = (APP_NAME, TOKEN)
    others_should_errback = ['CredentialsNotFound']

    @defer.inlineCallbacks
    def test_find_credentials_with_success_cb(self):
        """The credentials are asked and returned in a thread_execute call."""
        d = defer.Deferred()

        self.obj.find_credentials(APP_NAME, self.args,
                                  success_cb=d.callback, error_cb=d.errback)

        creds = yield d
        expected = yield self.factory.credentials.find_credentials()
        self.assertEqual(creds, expected)
        self.assert_recorder_called(self.factory, 'Credentials', APP_NAME)

    @defer.inlineCallbacks
    def test_find_credentials_with_error_cb(self):
        """If find_credentials fails, error_cb is called."""
        d = defer.Deferred()

        self.patch(self.factory.credentials, 'find_credentials',
                   lambda *a: defer.fail(SampleException('baz')))
        self.obj.find_credentials(APP_NAME, self.args,
                                  success_cb=d.errback, error_cb=d.callback)

        errdict = yield d
        self.assertEqual(errdict["errtype"], "SampleException")
        self.assert_recorder_called(self.factory, 'Credentials', APP_NAME)


class CredentialsManagementNotFindTestCase(CredentialsManagementFindTestCase):
    """Tests for the CredentialsManagement find method."""

    success_signal = 'CredentialsNotFound'
    success_result = (APP_NAME,)
    others_should_errback = ['CredentialsFound']

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementNotFindTestCase, self).setUp()
        self.call_method = getattr(self.obj, self.method)
        self.patch(self.factory.credentials, self.method,
                   lambda: defer.succeed({}))


class CredentialsManagementOpsTestCase(CredentialsManagementTestCase):
    """Tests for the CredentialsManagement login/register methods."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementOpsTestCase, self).setUp()
        self.args = dict(self.base_args)
        self.args[WINDOW_ID_KEY] = str(self.args[WINDOW_ID_KEY])
        self.cred_args = self.base_args.copy()

    def test_register(self):
        """The registration is correct."""
        self.obj.register(APP_NAME, self.args)
        self.assert_recorder_called(self.factory, 'Credentials',
                                    APP_NAME, **self.cred_args)

    def test_login(self):
        """The login is correct."""
        self.obj.login(APP_NAME, self.args)
        self.assert_recorder_called(self.factory, 'Credentials',
                                    APP_NAME, **self.cred_args)

    def test_login_email_password(self):
        """The login_email_password is correct."""
        self.args['email'] = EMAIL
        self.args['password'] = PASSWORD
        self.obj.login_email_password(APP_NAME, self.args)
        self.assert_recorder_called(self.factory, 'Credentials',
                                    APP_NAME, **self.cred_args)


class CredentialsManagementParamsTestCase(CredentialsManagementOpsTestCase):
    """Tests for the CredentialsManagement extra parameters handling."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementParamsTestCase, self).setUp()
        self.args['dummy'] = 'nothing useful'


class CredentialsManagementSignalsTestCase(CredentialsManagementTestCase):
    """Tests for the CredentialsManagement DBus signals."""

    def assert_signal_logged(self, signal, *args):
        """Check that signal info was properly logged."""
        signal(APP_NAME, *args)
        msgs = [self.obj.__class__.__name__,
                signal.__name__, repr(APP_NAME)]
        self.assertTrue(self.memento.check_info(*msgs))

    def test_credentials_found(self):
        """The CredentialsFound signal."""
        self.assert_signal_logged(self.obj.CredentialsFound, TOKEN)

        msg = 'credentials must not be logged (found %r in log).'
        for val in TOKEN.itervalues():
            self.assertFalse(self.memento.check_info(val), msg % val)

    def test_credentials_not_found(self):
        """The CredentialsNotFound signal."""
        self.assert_signal_logged(self.obj.CredentialsNotFound)

    def test_credentials_cleared(self):
        """The CredentialsCleared signal."""
        self.assert_signal_logged(self.obj.CredentialsCleared)

    def test_credentials_stored(self):
        """The CredentialsStored signal."""
        self.assert_signal_logged(self.obj.CredentialsStored)

    def test_credentials_error(self):
        """The CredentialsError signal."""
        error = {'error_message': 'failed!', 'detailed error': 'yadda yadda'}
        self.obj.CredentialsError(APP_NAME, error)
        msgs = (self.obj.__class__.__name__,
                self.obj.CredentialsError.__name__,
                repr(APP_NAME), repr(error))
        self.assertTrue(self.memento.check_error(*msgs))

    def test_authorization_denied(self):
        """The AuthorizationDenied signal."""
        self.assert_signal_logged(self.obj.AuthorizationDenied)


class UbuntuSSOServiceTestCase(BaseTestCase):
    """Test suite for the UbuntuSSOService class."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(UbuntuSSOServiceTestCase, self).setUp()
        self.proxy.sso_login = object()
        self.proxy.cred_manager = object()
        self.params = []
        self.patch(main, 'UbuntuSSOProxy',
                   lambda _: self.params.append(_) or self.proxy)
        self.obj = UbuntuSSOService()
        yield self.obj.start()

    def test_creation(self):
        """Creation paremeter is properly used."""
        self.assertEqual(self.params, [self.obj])

    def test_sso_login(self):
        """Attributes has the expected value."""
        self.assertIsInstance(self.obj.sso_login, SSOLogin)
        self.assertTrue(self.obj.sso_login.proxy is self.proxy.sso_login)

    def test_cred_manager(self):
        """Attributes has the expected value."""
        self.assertIsInstance(self.obj.cred_manager, CredentialsManagement)
        self.assertTrue(self.obj.cred_manager.proxy is self.proxy.cred_manager)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Perspective Broker IPC test cases."""

import logging

from collections import namedtuple

from twisted.internet import defer
from twisted.spread.pb import (
    DeadReferenceError,
    NoSuchMethod,
)
from ubuntuone.devtools.handlers import MementoHandler
from ubuntuone.devtools.testcases import skipIfOS
from ubuntuone.devtools.testcases.txsocketserver import (
    TidyUnixServer,
    TCPPbServerTestCase,
)

from ubuntu_kylin_sso.tests import TestCase
from ubuntu_kylin_sso.utils import ipc


TEST_SERVICE = 'foo-service'
TEST_CMDLINE = 'foo.bin'
TEST_SERVER_DESCRIPTION = 'tcp:40404:interface=127.0.0.1'
TEST_CLIENT_DESCRIPTION = 'tcp:host=127.0.0.1:port=40404'


class RandomException(Exception):
    """A random exception."""


class FakeActivationClient(object):
    """A fake ActivationClient."""

    def __init__(self, config):
        """Initialize this fake instance."""
        self.config = config

    def get_active_client_description(self):
        """Return the description where the pb server is running."""
        return defer.succeed(self.config.description.client)


class FakeActivationInstance(object):
    """A fake ActivationInstance."""

    def __init__(self, config):
        """Initialize the fake instance."""
        self.config = config

    def get_server_description(self):
        """Return the description where the pb server is running."""
        return defer.succeed(self.config.description.server)


class FakeDescriptionFactory(object):
    """A fake description factory."""

    def __init__(self, server_description, client_description):
        """Create a new instace."""
        self.server = server_description
        self.client = client_description


# Invalid name "callRemote", etc.
# pylint: disable=C0102, C0103


class FakeReactor(object):
    """A fake reactor."""

    def __init__(self):
        """Initialize this faker."""
        self.connection_class = namedtuple("Connection",
                                         "host port factory backlog")
        self.connections = []

    def connectTCP(self, host, port, factory, timeout=None, bindAddress=None):
        """Store the connected factory."""
        connection = self.connection_class(host, port, factory, None)
        self.connections.append(connection)

    def listenTCP(self, port, factory, interface=None, backlog=None):
        """Store the listenning factory."""
        connection = self.connection_class(interface, port, factory, backlog)
        self.connections.append(connection)


class FakeTCP4ClientEndpoint(object):
    """A fake tcp4 client."""

    def __init__(self, protocol):
        """Create a new instance."""
        self.protocol = protocol

    def connect(self, *args, **kwargs):
        """Return the client."""
        return defer.succeed(self.protocol)


class FakeRemoteClient(object):
    """A fake RemoteClient."""

    missing_signal = "missing"
    failing_signal = "failing"
    dead_remote = False
    random_exception = RandomException()

    def __init__(self, dead_remote=False):
        self.called = False
        self.dead_remote = dead_remote

    def callRemote(self, signal_name, *a, **kw):
        """Fake a call to a given remote method."""
        if self.dead_remote:
            raise DeadReferenceError("Calling Stale Broker")

        if signal_name == self.missing_signal:
            return defer.fail(NoSuchMethod())
        if signal_name == self.failing_signal:
            return defer.fail(self.random_exception)

        self.called = (a, kw)
        return defer.succeed(None)


class DummyRemoteService(ipc.RemoteService):
    """Represent a dummy IPC object."""

    remote_calls = ['foo', 'bar']
    next_result = None

    def foo(self):
        """Dummy foo."""
        self.Success(self.next_result)

    def bar(self, error):
        """Dummy bar."""
        self.NotSuccess(error)

    @ipc.signal
    def Success(self, param):
        """Fire a signal to notify a success."""

    @ipc.signal
    def NotSuccess(self, error):
        """Fire a signal to notify a not-success."""

    @ipc.signal
    def NoArgs(self):
        """Get no args passed."""

    @ipc.signal
    def JustArgs(self, *args):
        """Just get args."""

    @ipc.signal
    def JustKwargs(self, **kwargs):
        """Just get kwargs."""

    @ipc.signal
    def BothArgsAndKwargs(self, *args, **kwargs):
        """Both args and kwargs."""

# pylint: enable=C0102, C0103


class DummyService(ipc.BaseService):
    """Represent a dummy root service."""

    services = {'dummy': DummyRemoteService}
    name = 'Dummy Service'
    description = TEST_CLIENT_DESCRIPTION
    cmdline = 'yadda yo'


class DummyRemoteClient(ipc.RemoteClient):
    """Represent a dummy remote client."""

    call_remote_functions = DummyRemoteService.remote_calls
    signal_handlers = ['Success', 'NotSuccess']


class DummyClient(ipc.BaseClient):
    """Represent a dummy base client."""

    clients = {'dummy': DummyRemoteClient}
    service_name = DummyService.name
    service_port = TEST_SERVER_DESCRIPTION
    service_cmdline = DummyService.cmdline


class DummyDescription(object):
    """Return the descriptions accordingly."""

    def __init__(self, client, server):
        """Create a new instance."""
        self.client = client
        self.server = server


class BaseIPCTestCase(TCPPbServerTestCase, TestCase):
    """Set the ipc to a random port for this instance."""

    timeout = 5

    client_class = None  # the BaseClient instance
    service_class = None  # the BaseService instance

    remote_client_name = None  # the name of the remote client in the client
    remote_service_name = None  # the name of the remote service in the service

    method_mapping = []
    signal_mapping = []

    @defer.inlineCallbacks
    def setUp(self):
        # pylint complains about callables
        # pylint: disable=E1102
        yield super(BaseIPCTestCase, self).setUp()

        self.service = None
        self.client = None

        if self.service_class is not None:

            self.service = self.service_class()
            self.client = self.client_class()

            # patch server connection and client connection to ensure that
            # we have clean connections

            @defer.inlineCallbacks
            def server_listen(server_factory, service_name, activation_cmd,
                              description, reactor=None):
                """Connect to the local running service."""
                yield self.listen_server(self.service)
                defer.returnValue(self.listener)

            self.patch(ipc, 'server_listen', server_listen)

            @defer.inlineCallbacks
            def client_connect(client_factory, service_name,
                               activation_cmdline, description, reactor=None):
                """Connect the local running client."""
                yield self.connect_client()
                self.client.factory = self.client_factory
                defer.returnValue(self.connector)

            self.patch(ipc, 'client_connect', client_connect)

            # pylint: disable=E1102
            yield self.service.start()
            yield self.client.connect()

        # pylint: enable=E1102

    @property
    def remote_service(self):
        """Get the service named 'service_name'."""
        return getattr(self.service, self.remote_service_name)

    @property
    def remote_client(self):
        """Get the client named 'remote_client_name'."""
        return getattr(self.client, self.remote_client_name)

    @defer.inlineCallbacks
    def assert_method_called(self, service, method, result, *args, **kwargs):
        """Check that calling 'method(*args, **kwargs)' should query 'service'.

        The returned result from calling 'method(*args, **kwargs)' should be
        equal to the given parameter 'result'. If 'result' is a deferred, its
        result attribute will be used as expected result (ergo the deferred
        should be already called).

        """
        client = self.remote_client

        # hack to handle service methods returning a deferred with result
        if isinstance(result, defer.Deferred):
            real_result = result.result
        else:
            real_result = result

        self.patch(service, method, lambda *a, **kw: result)
        actual = yield client.call_method(method, *args, **kwargs)
        self.assertEqual(real_result, actual)
        self.assertEqual(service.called, {method: [(args, kwargs)]})

    def assert_remote_method(self, method_name, *args, **kwargs):
        """Assert that 'method_name' is a remote method.

        The parameters args and kwargs will be passed as such to the method
        itself, to exercise it.

        """
        self.assertIn(method_name, self.remote_service.remote_calls)
        method = getattr(self.remote_service, method_name)
        method(*args, **kwargs)

    def assert_remote_signal(self, signal_name, *args, **kwargs):
        """Assert that 'signal' is a remote signal.

        The parameters args and kwargs will be passed as such to the signal
        itself, to exercise it.

        """
        self.patch(self.remote_service, 'emit_signal', self._set_called)
        signal = getattr(self.remote_service, signal_name)
        signal(*args, **kwargs)

        expected = (signal_name,) + args
        self.assertEqual(self._called, (expected, kwargs))

    def test_remote_methods(self):
        """Check every method defined in self.method_mapping.

        Assert that every method is a remote method and that it has the
        expected signature.

        """
        for method, args, kwargs in self.method_mapping:
            self.assert_remote_method(method, *args, **kwargs)

    def test_remote_signals(self):
        """Check every signal defined in self.signal_mapping.

        Assert that every signal is a remote signal and that it has the
        expected signature.

        """
        for signal_name, args, kwargs in self.signal_mapping:
            self.assert_remote_signal(signal_name, *args, **kwargs)
# pylint: enable=E1102


class TCPListenConnectTestCase(BaseIPCTestCase):
    """Test suite for the server_listen and client_connect methods."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(TCPListenConnectTestCase, self).setUp()
        self.fake_reactor = FakeReactor()

    @defer.inlineCallbacks
    def test_server_listen(self):
        """Test the server_listen function."""
        self.patch(ipc, "ActivationInstance", FakeActivationInstance)

        description_factory = FakeDescriptionFactory(TEST_SERVER_DESCRIPTION,
                                                     TEST_CLIENT_DESCRIPTION)

        fake_factory = object()
        yield ipc.server_listen(fake_factory, TEST_SERVICE,
                                TEST_CMDLINE, description_factory,
                                reactor=self.fake_reactor)

        self.assertEqual(len(self.fake_reactor.connections), 1)
        conn = self.fake_reactor.connections[0]
        self.assertEqual(conn.factory, fake_factory)
        self.assertEqual(conn.host, ipc.LOCALHOST)

    @defer.inlineCallbacks
    def test_client_connect(self):
        """Test the client_connect function."""
        called = []
        self.patch(ipc, "ActivationClient", FakeActivationClient)

        protocol = 'protocol'
        client = FakeTCP4ClientEndpoint(protocol)

        def client_from_string(reactor, description):
            """Create a client from the given string."""
            called.append(('clientFromString', reactor, description))
            return client

        self.patch(ipc.endpoints, 'clientFromString', client_from_string)

        description_factory = FakeDescriptionFactory(TEST_SERVER_DESCRIPTION,
                                                     TEST_CLIENT_DESCRIPTION)

        fake_factory = object()
        returned_protocol = yield ipc.client_connect(fake_factory,
                                              TEST_SERVICE,
                                              TEST_CMDLINE,
                                              description_factory,
                                              reactor=self.fake_reactor)

        self.assertIn(('clientFromString', self.fake_reactor,
                          description_factory.client), called)
        self.assertEqual(protocol, returned_protocol)


@skipIfOS('win32', 'Unix domain sockets not supported on windows.')
class DomainListenConnectTestCase(TCPListenConnectTestCase):
    """Test suite for the server_listen and client_connect methods."""

    def get_server(self):
        """Return the server to be used to run the tests."""
        return TidyUnixServer()


class TCPDummyClientTestCase(BaseIPCTestCase):
    """Test the status client class."""

    client_class = DummyClient
    service_class = DummyService

    remote_client_name = remote_service_name = 'dummy'

    method_mapping = [
        ('foo', (), {}),
        ('bar', (object(),), {}),
    ]
    signal_mapping = [
        ('Success', ('test',), {}),
        ('NotSuccess', ('yadda',), {}),
        ('NoArgs', (), {}),
        ('JustArgs', (object(), 'foo'), {}),
        ('JustKwargs', (), {'foo': 'bar'}),
        ('BothArgsAndKwargs', ('zaraza', 8), {'foo': -42}),
    ]

    @defer.inlineCallbacks
    def test_deprecated_siganl_is_also_sent(self):
        """Old-style, deprecated signals handler are also called."""
        d1 = defer.Deferred()
        d2 = defer.Deferred()

        self.remote_service.next_result = 'yadda'

        # old, deprecated way
        self.remote_client.connect_to_signal('Success', d1.callback)
        self.remote_client.on_success_cb = d2.callback

        self.remote_client.foo()

        result = yield defer.gatherResults([d1, d2])

        self.assertEqual(result, ['yadda', 'yadda'])

    @defer.inlineCallbacks
    def test_register_to_signals(self):
        """Test the register_to_signals method."""
        yield self.remote_client.register_to_signals()
        self.addCleanup(self.remote_client.unregister_to_signals)

        for signal in self.remote_client.signal_handlers:
            self.assertIn(signal, self.service.dummy.clients_per_signal)

    @defer.inlineCallbacks
    def test_unregister_to_signals(self):
        """Test the register_to_signals method."""
        yield self.remote_client.register_to_signals()
        yield self.remote_client.unregister_to_signals()

        for signal in self.remote_client.signal_handlers:
            actual = len(self.remote_service.clients_per_signal[signal])
            self.assertEqual(0, actual)


@skipIfOS('win32', 'Unix domain sockets not supported on windows.')
class DomainDummyClientTestCase(TCPDummyClientTestCase):
    """Test the status client class."""

    def get_server(self):
        """Return the server to be used to run the tests."""
        return TidyUnixServer()


class RemoteMetaTestCase(TestCase):
    """Tests for the RemoteMeta metaclass."""

    def test_remote_calls_renamed(self):
        """The remote_calls are renamed."""
        test_token = object()

        # pylint: disable=W0232
        class TestClass(ipc.meta_base(ipc.RemoteMeta)):
            """A class for testing."""

            remote_calls = ['test_method']

            def test_method(self):
                """Fake call."""
                return test_token
        # pylint: enable=W0232

        tc = TestClass()
        self.assertEquals(tc.test_method(), test_token)
        # pylint: disable=E1101
        self.assertEquals(tc.remote_test_method(), test_token)
        # pylint: enable=E1101


class SignalBroadcasterTestCase(TestCase):
    """Test the signal broadcaster code."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(SignalBroadcasterTestCase, self).setUp()
        self.client = FakeRemoteClient()
        self.sb = ipc.SignalBroadcaster()

        self.memento = MementoHandler()
        ipc.logger.addHandler(self.memento)
        ipc.logger.setLevel(logging.DEBUG)
        self.addCleanup(ipc.logger.removeHandler, self.memento)

    def test_remote_register_to_signals(self):
        """Assert that the client was added."""
        signals = ["demo_signal1", "demo_signal2"]
        self.sb.remote_register_to_signals(self.client, signals)
        for signal in signals:
            clients = self.sb.clients_per_signal[signal]
            self.assertTrue(self.client in clients)

    def test_emit_signal(self):
        """Assert that the client method was called."""
        first = 1
        second = 2
        word = 'word'
        signal_name = 'on_test'

        self.client.callRemote(signal_name, first, second, word=word)

        signals = [signal_name]
        self.sb.remote_register_to_signals(self.client, signals)
        self.sb.emit_signal(signal_name, first, second, foo=word)

        self.assertEqual(self.client.called, ((first, second), dict(foo=word)))

    def test_emit_signal_dead_reference(self):
        """Test dead reference while emitting a signal."""
        sample_signal = "sample_signal"
        fake_remote_client = FakeRemoteClient(dead_remote=True)

        self.sb.remote_register_to_signals(fake_remote_client, [sample_signal])
        self.assertIn(fake_remote_client,
                      self.sb.clients_per_signal[sample_signal])

        self.sb.emit_signal(sample_signal)
        self.assertNotIn(fake_remote_client,
                         self.sb.clients_per_signal[sample_signal])

    def test_emit_signal_some_dead_some_not(self):
        """Test a clean reference after a dead one."""
        sample_signal = "sample_signal"
        fake_dead_remote = FakeRemoteClient(dead_remote=True)
        fake_alive_remote = FakeRemoteClient()

        self.sb.remote_register_to_signals(fake_dead_remote, [sample_signal])
        self.sb.remote_register_to_signals(fake_alive_remote, [sample_signal])
        self.sb.emit_signal(sample_signal)

        self.assertTrue(fake_alive_remote.called, "The alive must be called.")

    def test_emit_signal_ignore_missing_handlers(self):
        """A missing signal handler should just log a debug line."""
        fake_remote_client = FakeRemoteClient()

        signal = fake_remote_client.missing_signal
        signals = [signal]
        self.sb.remote_register_to_signals(fake_remote_client, signals)
        sb_clients = self.sb.clients_per_signal[signal]
        self.assertIn(fake_remote_client, sb_clients)
        self.sb.emit_signal(signal)

        expected = ipc.SignalBroadcaster.MSG_NO_SIGNAL_HANDLER % (
            signal,
            fake_remote_client,
        )
        self.assertTrue(self.memento.check_debug(*expected))

    def test_emit_signal_log_other_errors(self):
        """Other errors should be logged as warnings."""
        fake_remote_client = FakeRemoteClient()

        signal = fake_remote_client.failing_signal
        signals = [signal]
        self.sb.remote_register_to_signals(fake_remote_client, signals)
        sb_clients = self.sb.clients_per_signal[signal]
        self.assertIn(fake_remote_client, sb_clients)
        self.sb.emit_signal(signal)

        expected = ipc.SignalBroadcaster.MSG_COULD_NOT_EMIT_SIGNAL % (
            signal,
            fake_remote_client,
            fake_remote_client.random_exception,
        )
        self.assertTrue(self.memento.check_warning(*expected))


class FakeRootObject(object):
    """A fake root object."""

    def __init__(self, called, remote_obj):
        """Create a new instance."""
        self.called = called
        self.remote_obj = remote_obj

    # pylint: disable=C0103
    def callRemote(self, method_name):
        """A fake call remove method."""
        self.called.append(method_name)
        return defer.succeed(self.remote_obj)
    # pylint: enable=C0103


class FakeWorkingRemoteClient(object):
    """A fake remote client."""

    def __init__(self, called):
        """Create a new instance."""
        self.remote = None
        self.called = called

    def register_to_signals(self):
        """Register to signals."""
        self.called.append('register_to_signals')
        return defer.succeed(True)


class ReconnectTestCase(TestCase):
    """Test the reconnection when sso is dead."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(ReconnectTestCase, self).setUp()
        self.called = []
        self.remote_obj = 'remote'
        self.root_obj = FakeRootObject(self.called, self.remote_obj)

        def fake_get_root_object():
            """Fake getting the root object."""
            self.called.append('getRootObject')
            return defer.succeed(self.root_obj)

        def fake_client_connect(factory, service_name, cmd, description):
            """Fake the client connect."""
            self.called.append('client_connect')
            self.patch(factory, 'getRootObject', fake_get_root_object)
            return defer.succeed(True)

        self.patch(ipc, 'client_connect', fake_client_connect)

    @defer.inlineCallbacks
    def test_reconnect_method(self):
        """Test the execcution of the reconnect method."""
        clients = dict(first=FakeWorkingRemoteClient(self.called),
                       second=FakeWorkingRemoteClient(self.called))

        base_client = ipc.BaseClient()
        base_client.clients = clients
        for name, client in clients.items():
            setattr(base_client, name, client)

        yield base_client.reconnect()
        # assert that we did call the correct methods
        self.assertIn('client_connect', self.called)
        self.assertIn('getRootObject', self.called)

        for name in clients:
            self.assertIn('get_%s' % name, self.called)

        self.assertEqual(len(clients),
                self.called.count('register_to_signals'))
